#!/usr/bin/perl -w

# Linux Application Checker
# Results Page Module (tests_results.pl)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use Templates;
use Fcntl ':flock';
use Fcntl ':seek';

use Analysis; # utils/Analysis.pm

autoflush_on();

# Reads information about the tested application
sub read_results_info($) {
	my ($res_dir) = @_;

	my $app_name = "";
	if (open(FILE, "$res_dir/INFO")) {
		while (<FILE>) {
			if (m/Application: (.*)$/) {
				$app_name = $1;
				last;
			}
		}
		close(FILE);
	}
	my $res_dir_esc = $res_dir;
	$res_dir_esc =~ s/([ ()])/\\$1/g;
	my $host_platform = `cat $res_dir_esc/test_logs/host_platform`;
	my $lsb_version = `cat $res_dir_esc/test_logs/lsb_version`;
	$host_platform =~ s/[\r\n]//g;
	$lsb_version =~ s/[\r\n]//g;

	return ($app_name, $host_platform, $lsb_version);
}


if ($_POST{"remove"}) {
	if (open(FILE, "+<".$CONFIG{"RESULTS_DIR"}."/HISTORY")) {
		flock(FILE, LOCK_EX);
		while(<FILE>) {
			s/[\r\n]//g;
			next if ($_ eq "");
			my $path = $CONFIG{"RESULTS_DIR"}."/$_";
			if (-d $path) {
				if (defined($_POST{$_})) {
					$path =~ s/([ ()])/\\$1/g;
					system("rm -rf $path");
				}
				else {
					push @list, $_;
				}
			}
		}
		seek(FILE, 0, SEEK_SET);
		truncate(FILE, 0);
		print FILE join("\n", @list)."\n";
		flock(FILE, LOCK_UN);
		close(FILE);
	}
	else {
		print STDERR "Cannot open HISTORY file!\n";
	}
	print "HTTP/1.0 302 Moved\nLocation: tests_results.pl\n\n";
	exit;
}

print "HTTP/1.0 200 OK\n";
print "Content-type: text/html\n\n";

my $popup_html_header = <<HEREDOC;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en" dir="ltr">
<head>
	<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
	<title>Details</title>
    
	<style type="text/css">
		body {
			font: small sans-serif;
		}
		a {
			text-decoration: none;
		}
		a:hover {
			text-decoration: underline;
		}
		div.box {
			border: 1px solid #aaa;
			background-color: #f9f9f9;
			padding: 5pt;
			padding-left: 15pt;
		}
	</style>
</head>
<body>
HEREDOC

my $popup_html_footer = '</body></html>';

my @res_tab_ids = ('distrs', 'libs', 'ints', 'cert');
my %res_tab_names = (
	'distrs' => 'Distribution Compatibility',
	'libs' => 'Required Libraries',
	'ints'  => 'Required Interfaces',
	'cert' => 'LSB Certification'
);

if ($_GET{"details"}) {
	$_GET{'tab'} = 'distrs' if (!$_GET{'tab'});
	my $res_dir = $CONFIG{"RESULTS_DIR"}."/".$_GET{"details"};
	my ($app_name, $host_platform, $lsb_version) = read_results_info($res_dir);
	my $header_info_tabs = '';

	# If we selected one of the correct tabs, print the header.
	# if (defined($res_tab_names{$_GET{"tab"}}))  # DBG: TODO:
	{
		my $err = load_analysis_data($res_dir);

		if ($err) {
			# TODO: print a better error message
			print '<p>Error: '.$err.'</p>';
			exit;
		}
		$header_info_tabs = get_title( $app_name, $host_platform, $res_dir ).
							summary( $lsb_version ).
							html_results_tabs( $_GET{'tab'} );
	}

	if ($_GET{"tab"} eq 'distrs') {

		if ( $_GET{"id"} ) {
			my $Did = $_GET{"id"};
			my $show = $_GET{"show"};

			print $popup_html_header;

			print distr_details($Did, $show);

			print $popup_html_footer;

			exit 0;
		}

		# Print header
		print_header("App Analysis", "res-distrs");
		print $header_info_tabs;

		print D_table( $lsb_version );

		print_footer();
		exit 0;
	} # end of the 'distrs' tab
	elsif ($_GET{"tab"} eq 'libs') {
		if ( $_GET{"lib"} ) {
			my $soname = $_GET{"lib"};

			print $popup_html_header;

			print lib_details($soname, $lsb_version);

			print $popup_html_footer;

			exit 0;
		}

		## POSTed selection:
		# Struct: $distr_sel->{$Did} = $checked;
		my $distr_sel;

		if ( defined $_POST{"submitted"} ) {
			# User's selection
			for my $distr ( values %$Analysis::all_distrs ) {
				my $Did = $distr->{ID};
				my $control_name = "distr-0-".$Did;
				if ( $_POST{$control_name} ) {
					$distr_sel->{$Did} = 1;
				} else {
					$distr_sel->{$Did} = 0;
				}
			}
		} else {
			# Default values
			for my $distr ( values %$Analysis::all_distrs ) {
				my $Did = $distr->{ID};
				$distr_sel->{$Did} = 1;
			}
		}

		if ( $_GET{"list"} ) {
		# Library details
			my $soname = $_GET{"list"};

			if ( $_GET{"rel"} ) {
				my $reopen = href(tab_link('libs').'&list='.$soname);

				print $popup_html_header;

				print <<HEREDOC;
<script language="javascript" type="text/javascript">
// <![CDATA[
if ( opener && opener.document.hiddensel ) {
	op_form = opener.document.hiddensel;
	op_form.action = '$reopen';
	op_form.target = window.name;
	op_form.submit();
} else {
	document.location = '$reopen';
}
// ]]>
</script>
<noscript>javascript disabled</noscript>
HEREDOC
				print $popup_html_footer;
			}
			else {
				print $popup_html_header;

				print '<h2>Library: <a href="http://www.linuxfoundation.org/navigator/browse/rawlib.php?cmd=list-byrunname&amp;RLrunname='.html(href($soname)).'" target="_blank"><u>'.html($soname).'</u></a></h2>';

				# Presence in Distributions
				print '<h3>Presence in Distributions</h3>';

				print lib_distr_list($soname, $distr_sel);

				print $popup_html_footer;
			}

			exit 0;
		}

		# Print the view
		print_header("App Analysis", "res-libs");
		print $header_info_tabs;

		print L_table( $distr_sel, $lsb_version );

		print_footer();
		exit 0;
	}  # end of the 'libs' tab
	elsif ($_GET{"tab"} eq 'ints') {
		if ( $_GET{"int"} ) {
			print $popup_html_header;
			
			my ($Iname, $ver) = split " ", $_GET{"int"};
			print int_details( $Iname, $ver, $lsb_version );
			
			print $popup_html_footer;
			exit 0;
		}
		
		print_header("App Analysis", "res-libs");
		print $header_info_tabs;

		print I_table( $lsb_version );

		print_footer();
		exit 0;
	} # end of the 'ints' tab
	elsif ($_GET{"tab"} eq 'cert') {
		my $name = "test_log.html";
		print_header("App Analysis", "res-cert");
		if (open(FILE, $CONFIG{"RESULTS_DIR"}."/".$_GET{"details"}."/$name")) {
			my $transfer = 0;
			while (<FILE>) {
				if (m/<body>/) {
					s/^.*<body>\n?//i;
					$transfer = 1;
				}
				if ($transfer) {
					s/(<\/body>|<\/html>)\n?//i;
					if (m/<!-- Test Report Header -->/) {
						print $header_info_tabs;
						print <<DATA;
<form action="cert_integration.pl?start" method="post" name="integr_data" style="display:none">
<input type="hidden" name="file" value="$CONFIG{'RESULTS_DIR'}/$_GET{'details'}/$_GET{'details'}.tgz" />
</form>
<p style="margin-top: 1.5em;"><a href="cert_integration.pl?start" onclick="javascript:document.integr_data.submit(); return false"><img src="images/medal_small.jpg" width="28" height="28" alt="" /></a> If there are no errors, you can easily <a href="cert_integration.pl?start" onclick="document.integr_data.submit(); return false"><b>apply for certification</b></a> by uploading the results to the <a href="http://$cert_sys_host$cert_sys_base" target="_blank">LSB Certification System</a>.</p>
DATA
					}
					else {
						print $_;
					}
				}
			}
			close(FILE);
		}
		else {
			print "<b><font color=\"red\">Sorry, the report '".$_GET{"details"}."' cannot be opened!</font><br />\nReason: $!</b>\n";
		}
	} # end of the 'cert' tab
	else {
		print_header("App Analysis", "res-distrs");
		print "<b><font color=\"red\">Invalid tab selected!</font></b>\n";
	}
} # end of  if $_GET{'details'}
else {
	print_header("Test Results", "results");

	print "<h1>Test Results</h1>\n";

	print "<form method=\"post\" action=\"tests_results.pl\" name=\"results_list\">\n<table class=\"main\">\n";
	print "  <tr>\n    <td><input type=\"checkbox\" name=\"check_all\" id=\"check_all\" onclick=\"javascript:check_unckeck_all();\" /></td>\n";
	print "    <th style=\"width: 14em;\">Date/Time</th>\n";
	print "    <th style=\"width: 30em;\">Comment</th>\n  </tr>\n";
	my @list = ();
	if (open(FILE, "+<".$CONFIG{"RESULTS_DIR"}."/HISTORY")) {
		flock(FILE, LOCK_EX);
		while(<FILE>) {
			s/[\r\n]//g;
			next if ($_ eq "");
			push @list, $_ if (-d $CONFIG{"RESULTS_DIR"}."/$_");
		}
		seek(FILE, 0, SEEK_SET);
		truncate(FILE, 0);
		print FILE join("\n", @list)."\n";
		flock(FILE, LOCK_UN);
		close(FILE);
		@list = reverse(@list);
		foreach my $name (@list) {
			my $full_name = $CONFIG{"RESULTS_DIR"}."/$name";
			my $result = 'unknown';
			my $app_name = "";
			my @comment = ();
			my $reading_comment = 0;
			if (open(FILE, "$full_name/INFO")) {
				while (<FILE>) {
					s/[\r\n]//g;
					if ($reading_comment) {
						push @comment, $_;
					}
					else {
						if ($_ eq '') {
							$reading_comment = 1;
						}
						elsif (m/^Verdict: (\S+)/i) {
							if ($1 eq "Success") {
								$result = 'pass';
							}
							elsif ($1 eq "Warnings") {
								$result = 'warn';
							}
							elsif ($1 eq "Failed") {
								$result = 'fail';
							}
						}
						elsif (m/^Application: (.*)$/i) {
							$app_name = $1;
						}
					}
				}
				close(FILE);
			}
			if (open(FILE, "$full_name/INFO_LSB")) {
				while (<FILE>) {
					s/[\r\n]//g;
					if (m/^Verdict: (Success|Warnings)/i) {
						$result .= '-lsb';
					}
				}
				close(FILE);
			}
			my @extraction = ();
			my $jcomment = join("<br />", @comment);
			if ($jcomment eq "") {
				$jcomment = "&nbsp;";
			}
			my $show_name = $name;
			$show_name =~ s/^.*-([^\-]*)\-([^\-]*)\-([^\-]*)\-([^\-]*)h\-([^\-]*)m\-([^\-]*)s/$3.$2.$1&nbsp;$4:$5:$6/;
			my $upload_icon = "";
			if (! -f $CONFIG{"RESULTS_DIR"}."/$name/test_logs/ftp_upload") {
				$upload_icon = "&nbsp;<a href=\"upload_data.pl?details=$name\" title=\"Upload Data to the Linux Foundation\"><img src=\"images/upload_data.png\" alt=\"\" width=\"16\" height=\"16\" /></a>";
			}
			print <<DATA;
  <tr>
    <td rowspan="3"><input type='checkbox' name='$name' onclick='javascript:update_state();' /></td>
    <td rowspan="1" style="border-bottom: none;"><img src="images/environment-spacer.gif" width="1" height="16" alt="" /></td>
    <td rowspan="3" valign="top">$jcomment</td>
  </tr>
  <tr>
    <td valign="middle" style="border: none;">
      <table cellpadding="0" cellspacing="0" border="0"><tr>
        <td style="border: none; padding: 0px;"><img src="images/result-$result.png" alt="$result" width="25" height="41" /></td>
        <td style="border: none; padding: 0px;">&nbsp;<a href="tests_results.pl?details=$name">$show_name</a><br />&nbsp;Application: $app_name</td>
      </tr></table>
    </td>
  </tr>
  <tr>
    <td style="border: none; text-align: right; vertical-align: bottom;" width="5" height="5">
      <a href="tests_results.pl?details=$name&amp;tab=distrs" title="Show distribution compatibility report"><img src="images/res-distrs.png" alt="" width="16" height="16" /></a>&nbsp;<a href="tests_results.pl?details=$name&amp;tab=libs" title="Show required libraries"><img src="images/res-libs.png" alt="" width="16" height="16" /></a>&nbsp;<a href="tests_results.pl?details=$name&amp;tab=ints" title="Show required interfaces"><img src="images/res-ints.png" alt="" width="16" height="16" /></a>&nbsp;<a href="tests_results.pl?details=$name&amp;tab=cert" title="Show LSB compliance report"><img src="images/res-cert.png" alt="" width="16" height="16" /></a>$upload_icon
    </td>
  </tr>
DATA
		}
	}
	print <<DATA;
</table>
<br /><input type="submit" name="remove" id="remove_button" value="Remove Selected Entries" disabled="disabled" onclick='javascript:return confirm_remove();' />
</form>

<script language="javascript" type="text/javascript">
// <![CDATA[
function count_checked() {
	var num = 0;
	var form = document.results_list;
	for (var i=0; i<form.length; ++i) {
		if ((form[i].type.toLowerCase() == "checkbox") && (form[i].name != "check_all") && form[i].checked) {
			++num;
		}
	}
	return num;
}

function update_state() {
	var elem = document.getElementById("remove_button");
	if (elem) {
		var num = count_checked();
		elem.disabled = (num == 0);
	}
}

function check_unckeck(box, check) {
	temp = box.onchange;
	box.onchange = null;
	box.checked = check;
	box.onchange = temp;
}

function check_unckeck_all() {
	var elem = document.getElementById("check_all");
	if (elem) {
		var checked = elem.checked;
		var form = document.results_list;
		for (var i=0; i<form.length; ++i) {
			if ((form[i].type.toLowerCase() == "checkbox") && (form[i].name != "check_all"))
				check_unckeck(form[i], checked);
		}
		update_state();
	}
}

function confirm_remove() {
	var num = count_checked();
	if (num == 0) {
		alert("Nothing selected!");
		return false
	}
	else
		return confirm("Do you wish to delete " + num + " selected report(s)?");
}
// ]]>
</script>
DATA
} # end of the result list (result history)

print_footer();

exit 0;

#-----------------------------------------------------------------------

# Uses Analysis module to load all the analysis results.
sub load_analysis_data {
	my ($res_dir) = @_;
	
	local *fail = sub {
		my ($what) = @_;
		
		print_header("App Analysis", "res-distrs");
		
		my $log_file = "$SERVER_PARAM{'APP_DATA'}/log/app-checker-server.log.$SERVER_PARAM{'PORT'}";
		
		print '<span style="color:red">';
		print $what.'<br />';
		print 'Looks like the Analyzer has failed.<br />';
		print '<a href="/get.pl'.$SERVER_PARAM{'APP_DATA'}.'/results/'.$_GET{'details'}.'/test_logs/tests_output" target="_blank">View test log</a>';
		print '</span>';
		
		print_footer();
		
		exit 1;
	};
	
	# Load distribution list (contains distribution name, version, vendor, release year)
	Analysis::load_distrs( $res_dir."/Distributions" )
		or fail("Failed to read $res_dir/Distributions.");

	# TODO: check if a error returned; replace all "die()" everywhere
	
	Analysis::read_lib_dump( $res_dir."/libs" );

	Analysis::read_LibAttrs( $res_dir."/lib_attrs", $Analysis::libs );

	Analysis::read_libKB( $res_dir."/lib_kb" );

	Analysis::read_int_dump( $res_dir."/ints" );

	Analysis::read_IntAttrs( $res_dir."/int_attrs", $Analysis::ints );
	
	# Analysis::read_intKB( $res_dir."/int_kb" ); # TODO: <<<

	return;
}

# Customizable tab bar
sub html_results_tabs {
	my ($selected) = @_;

	my $nl = "\n";
	my $res = "";

	# Table start
	$res .= '<table cellpadding="0" cellspacing="0" border="0" width="100%"'
		.' style="white-space: nowrap; font-weight: bold; text-align: center; font-size: 120%; margin-bottom: 5pt;">'.$nl;
	$res .= '<tr>'.$nl;

	# Left spacer
	$res .= '<td style="width: 50px; border-bottom: solid 1px #166a98; height: 2em;">';
	$res .= '<img src="images/environment-spacer.gif" width="50" height="1" alt="" />';
	$res .= '</td>'.$nl;

	my $tabs_num = scalar(@res_tab_ids);

	foreach my $id (@res_tab_ids) {
		my $name = $res_tab_names{$id};
		my $link = tab_link($id);

		# Tab
		if ( $id eq $selected ) {
			$res .= '<td style="border: solid 1px #166a98;'
						.' border-bottom: none; background-color: #a0d5f1; color: #004060;">';
		} else {
			$res .= '<td class="res-tab-highlight" style="border: solid 1px #166a98;"'
				.' onclick="javascript:document.location=\''.html_arg($link).'\'">';
		}
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '<img src="images/res-'.$id.'.png" width="16" height="16" alt="" />&nbsp;';
		$name =~ s/\s/&nbsp;/g;
		$res .= $name;
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '</td>';

		# Spacer
		$res .= '<td style="width: 10px; border-bottom: solid 1px #166a98;">';
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '</td>'.$nl;
	}

	$res .= '<td style="border-bottom: solid 1px #166a98; width: 100%;">';
	$res .= '<img src="images/environment-spacer.gif" width="1" height="1" alt="" />';
	$res .= '</td>';

	$res .= '</tr>';

	# Bottom line
	$res .= '<tr>';
	$res .= '<td colspan="'.(2 * $tabs_num+2).'" style="background-color: #a0d5f1; height: 10px;">';
	$res .= '<img src="images/environment-spacer.gif" width="1" height="1" alt="" />';
	$res .= '</td>';
	$res .= '</tr>';

	$res .= '</table>'.$nl;

	return $res;
}

# Small function for easy formatting a distro name.
sub distr_name {
	my ($distr) = @_;

	return $distr->{NAME}.' '.$distr->{VER};
}

# Renders a control for selecting distribution subset
sub distr_selector {
	my ($distr_sel) = @_;

	my $nl = "\n";
	my $res = '';

	my $sel_distrs = scalar(grep {$_} values %$distr_sel);
	my $total_distrs = scalar(keys %$distr_sel);

	my %distrs = ();
	foreach my $did (keys %$distr_sel) {
		my $distr = $Analysis::all_distrs->{$did};
		$distrs{$did} = {
			'NAME'   => distr_name($distr),
			'YEAR'   => $distr->{'YEAR'},
			'FAMILY' => $distr->{'NAME'},
			'VENDOR' => $distr->{'VENDOR'}
		};
	}

	my @groups = (
		{
			'NAME'        => 'no',
			'DESCRIPTION' => '* No grouping *',
			'SORT_FUNC'   => undef,
			'ICON_FUNC'   => undef
		},
		{
			'NAME'        => 'FAMILY',
			'DESCRIPTION' => 'By distro family',
			'SORT_FUNC'   => sub {lc($a) cmp lc($b)},
			'ICON_FUNC'   => sub($) {'images/distros/'.lc($_[0]).'.png'}
		},
		{
			'NAME'        => 'VENDOR',
			'DESCRIPTION' => 'By vendor',
			'SORT_FUNC'   => sub {lc($a) cmp lc($b)},
			'ICON_FUNC'   => sub($) {'images/vendors/'.lc($_[0]).'.png'}
		},
		{
			'NAME'        => 'YEAR',
			'DESCRIPTION' => 'By release year',
			'SORT_FUNC'   => sub {$b <=> $a},
			'ICON_FUNC'   => undef
		}
	);
	my $num_groups = scalar(@groups);

	my @distr_groups = ();
	my @group_sel = ();
	foreach (my $i = 1; $i < $num_groups; ++$i) {
		my $group = $groups[$i];
		my $gname = $group->{'NAME'};
		$distr_groups[$i] = {};
		$group_sel[$i] = {};
		foreach my $did (keys %distrs) {
			my $group_value = $distrs{$did}->{$gname};
			if (defined($group_sel[$i]->{$group_value})) {
				if (!$distr_sel->{$did}) {
					$group_sel[$i]->{$group_value} = 0;
				}
			}
			else {
				$group_sel[$i]->{$group_value} = $distr_sel->{$did};
			}
			if (!defined($distr_groups[$i]->{$group_value})) {
				$distr_groups[$i]->{$group_value} = [];
			}
			push @{$distr_groups[$i]->{$group_value}}, $did;
		}
	}

	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
var num_distrs = $sel_distrs;
var num_groups = $num_groups;

var distros_checked = new Array();
var distro_to_group = new Array();
var group_to_distros = new Array();

HEREDOC

	foreach my $did (sort {$a <=> $b} keys %$distr_sel) {
		$res .= "distros_checked[$did] = $distr_sel->{$did};\n";
	}
	$res .= "\n";

	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= "// Group $i: $groups[$i]->{'NAME'}\ndistro_to_group[$i] = new Array();\n";
		foreach my $did (sort {$a <=> $b} keys %distrs) {
			my $group_value = $distrs{$did}->{$groups[$i]->{'NAME'}};
			$group_value =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			$res .= "distro_to_group[$i][$did] = '$group_value';\n";
		}
	}
	$res .= "\n";

	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= "// Group $i: $groups[$i]->{'NAME'}\ngroup_to_distros[$i] = new Array();\n";
		foreach my $group_value (sort {lc $a cmp lc $b} keys %{$distr_groups[$i]}) {
			my $arr = join(', ', @{$distr_groups[$i]->{$group_value}});
			$group_value =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			$res .= "group_to_distros[$i]['$group_value'] = [$arr];\n";
		}
	}
	$res .= "\n";

	$res .= <<HEREDOC;
function update_sel_distrs() {
	document.getElementById('sel-distrs').innerHTML = num_distrs;
}

function my_reset() {
	var inputs = document.getElementsByTagName('input');
	for (var i = 0; i < inputs.length; i++) {
		if ((inputs[i].type == 'checkbox') && ((inputs[i].id.indexOf('distr-') == 0) || (inputs[i].id.indexOf('group-') == 0))){
			inputs[i].checked = 'checked';
		}
	}
	for (var did in distros_checked) {
		distros_checked[did] = 1;
	}
	num_distrs = $total_distrs;
	update_sel_distrs();
}

function checkGroupCheckbox(group_idx, group_value) {
	// Do not check the 'no' grouping
	if (group_idx == 0)
		return;
	for (var i=0; i<group_to_distros[group_idx][group_value].length; ++i) {
		var did = group_to_distros[group_idx][group_value][i];
		if (!document.getElementById('distr-' + group_idx + '-' + did).checked) {
			document.getElementById('group-' + group_idx + '-' + group_value).checked = false;
			return;
		}
	}
	document.getElementById('group-' + group_idx + '-' + group_value).checked = true;
}

function onDistrToggle(elem) {
	distrToggle(elem.id.substr(elem.id.lastIndexOf('-') + 1), elem.checked, 0);
	update_sel_distrs();
}

function distrToggle(did, checked, not_check_group) {
	if (distros_checked[did] == checked)
		return;
	distros_checked[did] = checked;
	if (checked)
		++num_distrs;
	else
		--num_distrs;
	for (var i=0; i<num_groups; ++i) {
		document.getElementById('distr-' + i + '-' + did).checked = checked;
		if ((i > 0) && (i != not_check_group))
			checkGroupCheckbox(i, distro_to_group[i][did]);
	}
}

function onGroupToggle(elem, group_idx, group_value) {
	// Do not check the 'no' grouping
	if (group_idx == 0)
		return;
	var checked = elem.checked;
	for (var i=0; i<group_to_distros[group_idx][group_value].length; ++i) {
		var did = group_to_distros[group_idx][group_value][i];
		distrToggle(did, checked, group_idx);
	}
	update_sel_distrs();
}

function switch_grouping(id) {
	for (var i=0; i<num_groups; ++i) {
		var div = document.getElementById('distr-grouping-' + i);
		if (div)
			div.style.display = ((i == id) ? '' : 'none');
	}
}
// ]]>
</script>
HEREDOC

	# Toogler
	$res .= '<a href="" onclick="javascript:return toggle(\'dseldiv\');"><b><u>Select distributions&hellip;</u></b></a> (<b><span id="sel-distrs">'.$sel_distrs.'</span></b> of <b>'.$total_distrs.'</b> selected)';

	# Print selector
	$res .= '<div id="dseldiv" style="display: none;">';

	# Print group options
	$res .= <<DATA;
<table style="border: solid 1px gray; background: #e0e0e0; padding: 2px; margin: 0.5em 0;"><tr><td>
Group: <select name="distr-grouping" id="distr-grouping" onchange="javascript:switch_grouping(this.value);" style="width: 12.5em;">
DATA
	foreach (my $i = 0; $i < $num_groups; ++$i) {
		$res .= "  <option value=\"$i\"".(($i == 0) ? ' selected="selected"' : '').">$groups[$i]->{'DESCRIPTION'}</option>\n";
	}
	$res .= <<DATA;
</select>
</td></tr></table>
DATA

	# Print DIVs for every grouping
	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= '<div id="distr-grouping-'.$i.'" style="border: solid 1px; padding: 3pt; margin-bottom: 3pt; height: 200pt; width: 200pt; overflow: auto; display: none;">'."\n";
		foreach my $group_value (sort {$groups[$i]->{'SORT_FUNC'}()} keys %{$distr_groups[$i]}) {
			my $group_value_id = $group_value;
			$group_value_id =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			my $g_checked = ($group_sel[$i]->{$group_value} ? ' checked="checked"' : '');
			my $g_icon = (defined($groups[$i]->{'ICON_FUNC'}) ? '&nbsp;<img src="'.$groups[$i]->{'ICON_FUNC'}($group_value).'" width="16" height="16" alt="" />' : '');
			$res .= <<DATA;
  <input type="checkbox" name="group-$i-$group_value_id" id="group-$i-$group_value_id"$g_checked onclick="javascript:onGroupToggle(this, $i, '$group_value_id');" />$g_icon&nbsp;$group_value<br />
DATA
			foreach my $did (sort {lc($distrs{$a}->{'NAME'}) cmp lc($distrs{$b}->{'NAME'})} @{$distr_groups[$i]->{$group_value}}) {
				my $d_checked = ($distr_sel->{$did} ? ' checked="checked"' : '');
				my $d_icon = '&nbsp;<img src="images/distros/'.lc($distrs{$did}->{'FAMILY'}).'.png" width="16" height="16" alt="" />';
				$res .= <<DATA;
  &nbsp;&nbsp;&nbsp;&nbsp;<input type="checkbox" name="distr-$i-$did" id="distr-$i-$did"$d_checked onclick="javascript:onDistrToggle(this);" />$d_icon&nbsp;$distrs{$did}->{'NAME'}<br />
DATA
			}
		}
		$res .= "</div>\n";
	}

	# Print DIV with plain list
	$res .= <<DATA;
<form name="distrsel" id="distrsel" method="post" action="">
<div id="distr-grouping-0" style="border: solid 1px; padding: 3pt; margin-bottom: 3pt; height: 200pt; width: 200pt; overflow: auto;">
DATA
	foreach my $did (sort {lc($distrs{$a}->{'NAME'}) cmp lc($distrs{$b}->{'NAME'})} keys %distrs) {
		my $d_checked = ($distr_sel->{$did} ? ' checked="checked"' : '');
		my $d_icon = '&nbsp;<img src="images/distros/'.lc($distrs{$did}->{'FAMILY'}).'.png" width="16" height="16" alt="" />';
		$res .= <<DATA;
  <input type="checkbox" name="distr-0-$did" id="distr-0-$did"$d_checked onclick="javascript:onDistrToggle(this);" />$d_icon&nbsp;$distrs{$did}->{'NAME'}<br />
DATA
	}
	$res .= <<DATA;
</div>
<input type="submit" name="submitted" value="Apply" style="width: 7em;" />
 &nbsp; &nbsp; 
<input type="button" value="Reset" style="width: 7em;" onclick="javascript:return my_reset();" />
</form>
</div>
DATA

	# Print hidden form with selected distros
	$res .= '<form name="hiddensel" action="" method="post" style="display:none">'.$nl;
	foreach my $Did ( keys %$distr_sel ) {
		next if not $distr_sel->{$Did};

		my $control_name = "distr-0-".$Did;

		$res .= '<input type="hidden" name="'.$control_name.'" value="1" />'.$nl;
	}
	$res .= '<input type="hidden" name="submitted" value="1" />'.$nl;
	$res .= '</form>'.$nl;

	return $res;
}

sub parse_app_component {
	my ($s) = @_;
	my $disp_path = $s;
	
	$s =~ s/^.*\.rpm\|([^|]+)/RPM:$1/;
	$s =~ s/^.*\.(tar\.gz|tgz)\|([^|]+)/TGZ:$2/;
	$s =~ s/^PKG:.*\|([^|]+)/PKG:$1/;
	$s =~ s/\|//g;
	
	$disp_path =~ s/\|$//;
	$disp_path =~ s/\|/ => /g;
	
	return ($s, $disp_path);
}

sub no_binary_div {
	return <<HEREDOC;
<div style="background:#fec;border:1px solid black;padding:10pt;margin-top:20pt;text-align:center;font-size:large">This application contains no binary files (ELF executables or libraries).<br />Please see the LSB Certification tab for the results of script analysis (Perl, Python, Shell scripts).</div>
HEREDOC
}

# Library compatibility table
sub L_table {
	my ($distr_sel, $selected_lsb_ver) = @_;

	my $res = "";

	my $nl = "\n"; # Set this to "" to make spaghetti HTML code

	# Check if the table is not empty
	if ( !%$Analysis::libs ) {
		$res .= no_binary_div();
		return $res;
	}

	$res .= <<HEREDOC;
The table below lists all external libraries required by your Application (based on DT_NEEDED ELF section). LSB and Distribution presence statuses are provided for each library. You can adjust distributions of interest and the report will be updated appropriately.<br /><br />
HEREDOC

	$res .= distr_selector( $distr_sel );
	
	my $link = href(tab_link('libs'));
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id )
{
	var v = document.getElementById( id );
	if ( v ) {
		v.style.display = (v.style.display == 'none' ? '' : 'none');
	}
	return false;
}

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes,resizable=1'
		);
	w.moveTo(100,0);
	w.focus();

	return w;
}

function det_open( soname )
{
	popup_win( '$link' + '&lib=' + soname );

	return false;
}

function lib_distrs_open( soname )
{
	popup_win( '$link' + '&rel=1&list=' + soname );

	return false;
}
// ]]>
</script>
HEREDOC
	
	my $filtered = scalar grep {!$_} values %$distr_sel;

	my $table = "";
	# Table Header
	my $table_open = '<table id="l_table" class="wikitable" width="100%">';
	$table .= $table_open.$nl;
	$table .= '<tr>'
		.'<th>Library</th>'
		.'<th>Dependent<br />Application Components</th>'
		.'<th>Interfaces<br />Actually Used</th>'
		.'<th>Presence in<br />'.($filtered ? 'Selected ' : '').'Distributions</th>'
		.'<th id="lsb_col_header">In LSB '.$selected_lsb_ver.'?</th>'
		.'<th>More Info</th>'
		.'</tr>'.$nl;

	my $libs = $Analysis::libs;
	my $libs_num = scalar keys %$libs;
	
	my $filt_show = {};
	my $filt_component = {};

	my $row_n = 0; # Counter for generating row-specific IDs

	# Table Body
	foreach my $soname ( sort {lc $a cmp lc $b} keys %$libs ) {
		my $lib_obj = $libs->{$soname};

		$row_n++;
		my $row = "";
		
		$row .= '<tr>'.$nl;

		## Library
		$row .= '<td><b><a href="http://www.linuxfoundation.org/navigator/browse/rawlib.php?cmd=list-byrunname&amp;RLrunname='.html(href($soname)).'" target="_blank">'.html($soname).'</a></b></td>'.$nl;

		# App Modules: preparation
		my $app_modules = "";
		my $app_comp_n = 0;
		if ( $lib_obj->{APP_MODULES} ) {
			foreach my $component ( sort @{$lib_obj->{APP_MODULES}} ) {
				my ($disp_name, $disp_path) = parse_app_component($component);
				
				$app_modules .= (($app_modules eq '') ? '' : '<br />');
				$app_modules .= '<span title="'.html_arg($disp_path).'">'.html($disp_name).'</span>';
				
				$filt_component->{$disp_path}{$row_n} = 1;
				
				$app_comp_n++;
			}
		}

		## App modules expandable list
		$row .= '<td>';
		my $div_id = "cl".$row_n;
		my $plu = ( $app_comp_n > 1 ? 's' : '' );
		$row .= '<a href="" onclick="return toggle(\''.$div_id.'\')">'
					.$app_comp_n.' component'.$plu.' (&hellip;)'
				.'</a>';
		$row .= '<div id="'.$div_id.'" style="display:none">';
		$row .= $app_modules;
		$row .= '</div>';
		$row .= '</td>'.$nl;

		# Interfaces used
		my $ints_used = ($lib_obj->{INTS_USED} or 0);
		$row .= '<td align="center">';
		if ( $ints_used ) {
			my $link = href( tab_link('ints').'&show_lib='.$soname );
			$row .= '<a href="'.html($link).'">'.$ints_used.'</a>';
		} else {
			$row .= '<span style="color:red"><b>'.$ints_used.'</b></span>';
			$filt_show->{'unneeded'}{$row_n} = 1;
		}
		$row .= '</td>'.$nl;

		## Presence in distributions

		my $total = scalar( grep {$_} values %$distr_sel );
		my $present = 0;
		if ( $lib_obj->{DISTRS} ) {
			$present = scalar( grep { $distr_sel->{$_} && $lib_obj->{DISTRS}{$_} } keys %$distr_sel );
		}
		my $approved = ($lib_obj->{APPROVED} or 0);

		my $bgcolor;
		if ( $present == $total ) {
			$bgcolor = "#CCFFCC"; # green
		} elsif ( $present > 0 ) {
			$bgcolor = "#FEFECC"; # yellow
			$filt_show->{'problems'}{$row_n} = 1;
		} else {
			$bgcolor = "#FFCCCC"; # red
			$filt_show->{'problems'}{$row_n} = 1;
		}

		$row .= '<td align="right" style="background:'.$bgcolor.'">';
		my $details_link = ' (<a href="'.html(href(tab_link('libs').'&list='.$soname)).'"'
			.' onclick="return lib_distrs_open(\''.html(href($soname)).'\')">list&hellip;</a>)';
		if ( $approved or $present == $total ) {
			$row .= $present." of ".$total;
			$row .= $details_link;
		} elsif ( $present > 0 ) {
			$row .= $present." known distro".($present > 1 ? 's' : '');
			$row .= $details_link;
			$filt_show->{'unapproved'}{$row_n} = 1;
		} else {
			$row .= 'Unknown library';
			# Tooltip:
			$row .= '&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, \'libs_ints_unknown\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
			$filt_show->{'unapproved'}{$row_n} = 1;
		}
		$row .= '</td>'.$nl;


		## LSB
		$row .= '<td align="center">';

		$row .= '<div>';

		my ($lsb_status, $lsb_status_ver) = lsb_status($lib_obj, $selected_lsb_ver);
		$status_ver = '' if not defined $status_ver;
		
		my $color = undef;

		if ( $lsb_status eq "NOT" || $lsb_status eq "WITHDRAWN" ) {
			$row .= '<span style="color:red">No</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "CANDIDATE" ) {
			$row .= '<span style="color:#907000">Candidate for '.$lsb_status_ver.'</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "FUTURE" ) {
			$row .= '<span style="color:#907000">In LSB '.$lsb_status_ver.'</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "OPTIONAL" ) {
			$row .= '<span style="color:blue">Optional</span>';
		}
		elsif ( $lsb_status eq "IN" ) {
			$row .= '<span style="color:green">Yes</span>';
		}
		elsif ( $lsb_status eq "DEPRECATED" ) {
			$row .= '<span style="color:blue">Deprecated</span>';
		}

		$row .= '</div>';

		$row .= '</td>'.$nl;


		## More Info

		$row .= '<td>';
		my $info_link = href( tab_link('libs').'&rel=1&lib='.$soname );
		my $info_onclick = 'onclick="return det_open(\''.html(href($soname)).'\')"';
		my $txt = 'Details&hellip;';
		if ( $lib_obj->{ATTR} ) {
			$txt = '<b>More&nbsp;Info&hellip;</b>';
		}
		if ( $Analysis::libKBDistr->{$soname} ) {
			$txt = '<b>Known&nbsp;Issues&hellip;</b>';
		}
		$row .= '<a href="'.html($info_link).'" '.$info_onclick.'>'.$txt.'</a>';
		$row .= '</td>'.$nl;

		$row .= '</tr>'.$nl;
		
		$table .= $row;
	}

	$table .= '</table>'.$nl;

	# Filters before the table
	$res .= '<table><tr>';
	
	# "Component" filter
	$res .= '<td><table class="bartable">'
  		.'<tr>'
  		.'<td align="right"><b>Component:</b></td>'
    	.'<td align="right">'
      	.'<select name="com_select" id="com_select" onchange="return apply_filter();" style="width: 12.5em;">'
        .'<option value="" selected="selected">All</option>';
		foreach my $component ( sort {lc $a cmp lc $b} keys %$filt_component ) {
			$res .= '<option value="'.html_arg($component).'">'.html($component).'</option>';
		}
    $res .= '</select>'
    	.'</td>'
  		.'</tr>'
		.'</table></td>';

	# "Show" filter
	$res .= '<td><table class="bartable">'
  		.'<tr>'
  		.'<td align="right"><b>Show:</b></td>'
    	.'<td align="right">'
      	.'<select name="filt" id="filt" onchange="return apply_filter();" style="width: 12.5em;">'
        .'<option value="" selected="selected">All</option>'
        .'<option value="problems">Compatibility problems</option>'
        .'<option value="unneeded">Unneeded libraries</option>'
        .'<option value="unapproved">Uncommon libraries</option>'
        .'<option value="non-lsb">Non-LSB</option>'
      	.'</select>'
    	.'</td>'
  		.'</tr>'
		.'</table></td>';
	
	# Filter status
	$res .= '<td id="lines_count" style="padding-left:20pt">'.$libs_num.' librar'.($libs_num!=1?'ies':'y').' shown.</td>';
	$res .= '<td>[<a href="" onclick="javascript:document.getElementById(\'filt\').value=\'\'; document.getElementById(\'com_select\').value=\'\'; return apply_filter();">Reset all filters</a>]</td>';
	$res .= '</tr></table>';
	
	## Here the Table goes
	$res .= '<div id="wrapper">'.$nl;
	$res .= $table;
	$res .= '</div>'.$nl;
	##
	
	# Javascript stuff
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
var wrapper = document.getElementById( 'wrapper' );
var table = document.getElementById( 'l_table' );
var lcp = document.getElementById( 'lines_count' );

var l = new Array();
var table_lines = table.getElementsByTagName('tr');
for ( var i=0; i<table_lines.length; i++ ) {
	l[i] = '<tr>' + table_lines[i].innerHTML + '</tr>';
}
HEREDOC

	# javascript: filt_show
	$res .= 'var filt_show = Array();'.$nl;
	foreach my $what ( keys %$filt_show ) {
		$res .= 'filt_show[\''.$what.'\']=new Array(0';
		foreach my $row_n ( keys %{$filt_show->{$what}} ) {
			$res .= ','.$row_n;
		}
		$res .= ');'.$nl;
	}

	# javascript: filt_com
	$res .= 'var filt_com = Array();'.$nl;
	foreach my $component ( keys %$filt_component ) {
		$res .= 'filt_com[\''.$component.'\']=new Array(0';
		foreach my $row_n ( keys %{$filt_component->{$component}} ) {
			$res .= ','.$row_n;
		}
		$res .= ');'.$nl;
	}
	
	$res .= <<HEREDOC;

// Convert filt_show into a hash map
for ( var what in filt_show ) {
	var hash = new Array();
	for ( var row_n in filt_show[what] ) {
		hash[filt_show[what][row_n]] = 1;
	}
	filt_show[what] = hash;
}

// Convert filt_com into a hash map
for ( var com in filt_com ) {
	var hash = new Array();
	for ( var row_n in filt_com[com] ) {
		hash[filt_com[com][row_n]] = 1;
	}
	filt_com[com] = hash;
}

function highlight_select( obj )
{
	if ( obj.options ) {
		for ( var i in obj.options ) {
			if ( obj.options[i] && obj.options[i].style ) {
				obj.options[i].style.backgroundColor = '#fff';
			}
		}
		if ( obj.selectedIndex ) {
			obj.options[obj.selectedIndex].style.backgroundColor = '#AAFFAA';
		}
	}
}

function apply_filter()
{
	var filt_obj = document.getElementById('filt');
	var filt = filt_obj.value;
	highlight_select(filt_obj);
	
	var com_select_obj = document.getElementById('com_select');
	var com_select = com_select_obj.value;
	highlight_select(com_select_obj);

	var new_l_table = l[0]; // header line
	var shown = 0;
	
	for ( var i=1; i < l.length; i++ ) {
		if ( com_select && (!filt_com[com_select] || !filt_com[com_select][i]) ) continue;
		if ( filt && (!filt_show[filt] || !filt_show[filt][i]) ) continue;
		new_l_table += l[i];
		shown++;
	}
	
	var total = l.length - 1;
	lcp.innerHTML = shown + (shown != total ? ' of '+total : '') +' librar'+(total!=1 ? 'ies' : 'y')+' shown.';
	
	wrapper.innerHTML = '$table_open' + new_l_table + '</table>';
	return false;
}

// ]]>
</script>
HEREDOC

	if ( $_GET{'show'} ) {
		my $filt_value = $_GET{'show'};
		
		$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
HEREDOC
		if ( $filt_value ) {
			$res .= <<HEREDOC;
	var filt = document.getElementById('filt');
	filt.value = '$filt_value';
HEREDOC
		}
		
		$res .= <<HEREDOC;
	apply_filter();
// ]]>
</script>
HEREDOC
	}
	
	return $res;
}

# Returns LSB status of the library and LSB version related to this status
# Values: NOT | CANDIDATE | IN | OPTIONAL | DEPRECATED | WITHDRAWN
sub lsb_status {
	my ($obj, $lsb_ver) = @_;

	return ("NOT") if !$obj;

	my %events = ();
	
	if ( $obj->{APPEAREDIN} ) {
		$events{$_} = 'APPEARED' foreach ( split /,\s*/, $obj->{APPEAREDIN} );
	}
	if ( $obj->{DEPRECATEDSINCE} ) {
		$events{$_} = 'DEPRECATED' foreach ( split /,\s*/, $obj->{DEPRECATEDSINCE} );
	}
	if ( $obj->{WITHDRAWNIN} ) {
		$events{$_} = 'WITHDRAWN' foreach ( split /,\s*/, $obj->{WITHDRAWNIN} );
	}
	
	my $cur_state = 'NOT';
	my $last_v = '';
	foreach $v ( sort keys %events ) {
		last if $v gt $lsb_ver;

		$cur_state = $events{$v};
		$last_v = $v;
	}
	
	if ( $cur_state eq 'NOT' ) {
		if ( $obj->{APPEAREDIN} ) {
			my $first_appear = (sort split /,\s*/, $obj->{APPEAREDIN})[0];
			return ("FUTURE", $first_appear);
		}
		elsif ( $obj->{CANDIDATEFOR} ) {
			return ("CANDIDATE", $obj->{CANDIDATEFOR});
		}
		else {
			return "NOT";
		}
	}
	elsif ( $cur_state eq 'APPEARED' ) {
		if ( $obj->{MANDATORYSINCE} && $obj->{MANDATORYSINCE} le $lsb_ver ) {
			return ("IN", $last_v);
		} else {
			return ("OPTIONAL", $last_v);
		}
	}
	elsif ( $cur_state eq 'DEPRECATED' ) {
		return ("DEPRECATED", $last_v);
	}
	elsif ( $cur_state eq 'WITHDRAWN' ) {
		return ("WITHDRAWN", $last_v);
	}
}

sub int_lsb_status {
	my ($int_obj, $selected_lsb_ver) = @_;
	
	my ($lsb_status, $lsb_status_ver) = lsb_status($int_obj, $selected_lsb_ver);

	if ( $lsb_status eq "IN" ) {
		# Check library status
		my ($first_lib) = @{$int_obj->{LIBRARIES}};
		if ( $first_lib =~ /^(.*)\.so/ && $int_obj->{LSB_LIBRARY}
			&& lc $int_obj->{LSB_LIBRARY} eq lc $1 )
		{
			my $lib = $Analysis::libs->{$first_lib};
			if ( $lib ) {
				if ( $lib->{DEPRECATEDSINCE} && $lib->{DEPRECATEDSINCE} le $selected_lsb_ver ) {
					$lsb_status = 'DEPRECATED';
					$lsb_status_ver = $lib->{DEPRECATEDSINCE};
				}
				if ( $lib->{WITHDRAWNIN} && $lib->{WITHDRAWNIN} le $selected_lsb_ver ) {
					$lsb_status = 'WITHDRAWN';
					$lsb_status_ver = $lib->{WITHDRAWNIN};
				}
			}
		}
	}
	
	return ($lsb_status, $lsb_status_ver);
}

# Contents of a popup window with details on a library presence
# in a selected distribution subset.
sub lib_distr_list {
	my ($soname, $distr_sel) = @_;

	my $nl = "\n";
	my $res = "";

	my $lib = $Analysis::libs->{$soname};

	my $filtered = 0;

	my %present = ();
	my %missed = ();
	foreach my $Did ( keys %$distr_sel ) {
		if ( not $distr_sel->{$Did} ) {
			$filtered = 1;
			next; # Skip unselected
		}

		my $distr_name = distr_name( $Analysis::all_distrs->{$Did} );
		if ( $lib->{DISTRS}{$Did} ) {
			$present{$distr_name} = $Did;
		} else {
			$missed{$distr_name} = $Did;
		}
	}

	my $approved = ($lib->{APPROVED} or 0);

	my $present_num = scalar(keys %present);
	my $missed_num = scalar(keys %missed);
	if ( $present_num > 0 ) {
		$res .= '<p>';
		$res .= 'Present in <span style="color:green">'.$present_num.'</span>';
		if ( $approved or $missed_num == 0) {
			$res .= ' and missing in <span style="color:red">'.$missed_num.'</span>'
		} else {
			$res .= ' and nothing is known about <span style="color:#907000">'.$missed_num.'</span>'; # yellow
		}
		if ( $filtered ) {
			$res .= ' of '.($present_num + $missed_num).' selected distributions.';
		} else {
			$res .= ' of total '.($present_num + $missed_num).' distributions.';
		}
		$res .= '</p>'.$nl;
	} else {
		if ( $approved ) {
			$res .= '<p>This library is missing in all the '.($present_num + $missed_num);
			if ( $filtered ) {
				$res .= ' selected distributions.</p>'.$nl;
			} else {
				$res .= ' analyzed distributions.</p>'.$nl;
			}
		} else {
			$res .= '<p>There is no information about this library in the database.</p>'.$nl;
		}
	}

	local *distr_list = sub {
		my ($hash, $color, $color_alt) = @_;

		my $res = "";

		$res .= '<ul>'.$nl; # Green list
		for my $distr_name ( sort {lc $a cmp lc $b} keys %$hash ) {
			my $Did = $hash->{$distr_name};
			my $alt = 0;
			if ( $Did < 0 ) {
				$alt = 1;
				$Did = -$Did;
			}
			my $link = href( tab_link('distrs').'&id='.$Did );
			
			if ( !$alt ) {
				$res .= '<li><a style="color:'.$color.'" href="'.html($link).'">'
							.html($distr_name).'</a></li>'.$nl;
			} else {
				$res .= '<li><a style="color:'.$color_alt.'" href="'.html($link).'">'
							.html($distr_name).'</a>'
						.' &nbsp; <span style="color:'.$color.'">*</span>'
						.'</li>'.$nl;
			}
		}
		$res .= '</ul>'.$nl;

		return $res;
	};

	if ( %present ) {
		$res .= '<h4>Present in:</h4>'.$nl;

		$res .= distr_list(\%present, 'green');
	}

	if ( %missed ) {
		if ( $approved ) {
			$res .= '<h4>Missing in:</h4>'.$nl;

			$res .= distr_list(\%missed, 'red', '#907000'); # yellow
		} else {
			$res .= '<h4>No information for:</h4>'.$nl;

			$res .= distr_list(\%missed, '#907000'); # yellow
		}
	}

	return $res;
}

# Contents of a details popup window for Libraries table
sub lib_details {
	my ($soname, $selected_lsb_ver) = @_;

	my $nl = "\n";
	my $res = "";

	my $lib = $Analysis::libs->{$soname};
	
	if ( !$lib ) {
		$res .= "Error: there is no such library: '".$soname."'";
		return $res;
	}
	
	my $soname_link = $soname;
	$soname_link =~ s/\+/%2B/g;

	$res .= '<h2>Library: <a href="http://www.linuxfoundation.org/navigator/browse/rawlib.php?cmd=list-byrunname&amp;RLrunname='.$soname_link.'" target="_blank"><u>'.$soname.'</u></a></h2>';

	$res .= '<hr/>'.$nl;

	# Check Lib Knowledge Base
	my $kb_info = lib_KB_info($soname);
	if ( $kb_info ) {
		$res .= '<h3>Known compatibility issues:</h3>'.$nl;
		$res .= '<p>'.$kb_info.'</p>'.$nl;

		$res .= '<hr />'.$nl;
	}

	# Check LibAttr
	if ( $lib->{ATTR} ) {
		foreach my $type ( sort keys %{$lib->{ATTR}} ) {
			my $attr = "";
			my $bgcolor = '#fec';
			if ( $type eq "Recommendation" ) {
				$attr .= '<h3>LSB Recommendation</h3>';
				$attr .= $lib->{ATTR}->{$type};
			}
			elsif ( $type eq "StaticLinkage" ) {
				$attr .= '<h3>May be linked statically</h3>';
				$attr .= $lib->{ATTR}->{$type};
			}
			elsif ( $type eq "LSB Rejection" ) {
				$attr .= '<h3>Rejection</h3>';
				$attr .= $lib->{ATTR}->{$type};
			}
			else { # Custom
				$attr .= $lib->{ATTR}->{$type};
			}
			$attr =~ s/^\s+//; # remove unnecessary spaces
			$attr =~ s/\s+$//; #
			$attr .= "." if $attr =~ /[^\s\.]\s*$/;
			
			$res .= '<div class="box" style="background:'.$bgcolor.'">'.$attr.'</div>';
		}

		$res .= '<hr />'.$nl;
	}

	# LSB Status
	$res .= '<h3>LSB Status:</h3>'.$nl;

	my ($lsb_status, $lsb_status_ver) = lsb_status($lib, $selected_lsb_ver);

	if ( $lsb_status eq "NOT" ) {
		$res .= '<div style="color:red">Never been in LSB.</div>'.$nl;
	}
	elsif ( $lsb_status eq "FUTURE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. (Appeared in LSB '.$lsb_status_ver.')</div>'.$nl;
	}
	elsif ( $lsb_status eq "CANDIDATE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. Candidate for LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "OPTIONAL" ) {
		$res .= '<div style="color:blue">Optional since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "IN" ) {
		$res .= '<div style="color:green">Mandatory since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "DEPRECATED" ) {
		$res .= '<div style="color:blue">Deprecated since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "WITHDRAWN" ) {
		$res .= '<div style="color:red">Withdrawn since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}

	$res .= '<hr />'.$nl;

	# Presence in Distributions
	$res .= '<h3>Presence in Distributions</h3>'.$nl;

	$res .= lib_distr_list($soname, $Analysis::all_distrs);

	$res .= '<hr />'.$nl;

	return $res;
}

# Returns information from the Knowledge Base for specified library and
#  distribution set.
sub lib_KB_info {
	my ($soname) = @_;

	my $nl = "\n";
	my $res = "";

	my $problems = $Analysis::libKBDistr->{$soname};

	return undef if not $problems;

	foreach my $LKBid ( sort {$a <=> $b} keys %$problems ) {
		my $text = $Analysis::libKB->[$LKBid];
		$text =~ s/\n/<br\/>\n/g;

		$res .= '<div>';
		$res .= "<p>";
		$res .= html($text);
		$res .= "</p>".$nl;

		my $grouped = $problems->{$LKBid};

		foreach my $i ( sort {$a <=> $b} keys %$grouped ) {
			$res .= '<div><b>Group '.$i.':</b> ';

			# Print IDs of distributions from this group
			$res .= join ", ", map {distr_name($Analysis::all_distrs->{$_}) } @{$grouped->{$i}};

			$res .= '.</div>'.$nl;
		}

		$res .= '</div>'
	}

	return $res;
}

# Distribution compatibility table
sub D_table {
	my ($selected_lsb_ver) = @_;

	if ( !%$Analysis::libs ) {
		$res .= no_binary_div();
		return $res;
	}

	my $all_distrs = $Analysis::all_distrs;
	#my $libs = $Analysis::libs;

	# Distro names
	my %distr_names = ();
	$distr_names{$_} = distr_name($all_distrs->{$_}) for keys %$all_distrs;
	
	# Calc distro-libs statistics
	my %lib_statistics = (); # $lib_statistics{$Did}{OK|UNNEEDED|UNKNOWN|MISSED} = $count;
	my %int_statistics = (); # $int_statistics{$Did}{OK|UNKNOWN|MISSED} = $count;
	my %distr_status = (); # $distr_status{$Did} = OK|UNNEEDED|UNKNOWN|MISSED;
	
	foreach my $Did ( keys %$all_distrs ) {
		($distr_status{$Did}, $lib_statistics{$Did}, $int_statistics{$Did}) = Analysis::distr_status($Did);
	}

	my $nl = "\n";
	my $res = "";

	my $link = href(tab_link('distrs'));
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes'
		);
	w.moveTo(100,0);
	w.focus();

	return w;
}

function det_open( id, show )
{
	popup_win( '$link' + '&id=' + id + '&show=' + show );

	return false;
}
// ]]>
</script>
HEREDOC

	$res .= <<DATA;
<div id="distro-details" style="position: absolute; border: 1px #aaaaaa solid; border-left: none; background: white; padding: 0px 2px; vertical-align: bottom; display: none;">
</div>

The table below shows the compatibility status of your application with the distributions analyzed by the Linux Foundation. Your Application will run on the "<font color="green">green</font>" distributions without loader problems. Compatibility with the "<font color="#c0a000">yellow</font>" distributions can be easily achieved by excluding unneeded libraries from the dependencies of your application. Making the Application compatible with the "<font color="red">red</font>" distributions may require more effort to avoid using missing libraries/interfaces or by supplying them as a part of your application package. Please note that functional correctness is not guaranteed by this analysis.
DATA

	my $ext_table_display = ( $_GET{'expand'} ? '' : 'none' );
	my $ext_table = <<DATA;
<table class="wikitable" cellpadding="0" id="d_table_ext" style="display: $ext_table_display; margin: 0; border-left: none;">
<tr><th colspan="2" style="border-left: none;">Missing&nbsp;Libraries</th><th colspan="2" style="">Missing&nbsp;Interfaces</th></tr>
<tr>
  <th style="width: 5em; border-left: none;"><table cellpadding="0" cellspacing="0" border="0" width="100%"><tr><td align="center" style="border: none; white-space: nowrap;">Common&nbsp;<a href="tests_help.pl#libs_ints_common" onclick="javascript:show_hide_tooltip(this, 'libs_ints_common', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></td></tr></table></th>
  <th style="width: 5em; white-space: nowrap;">Unknown&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, 'libs_ints_unknown', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
  <th style="width: 5em; white-space: nowrap;">Common&nbsp;<a href="tests_help.pl#libs_ints_common" onclick="javascript:show_hide_tooltip(this, 'libs_ints_common', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
  <th style="width: 5em; white-space: nowrap;">Unknown&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, 'libs_ints_unknown', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
</tr>
DATA

	my $general_table = <<DATA;
<table class="wikitable" id="d_table" style="margin: 0;">
<tr><th colspan="2">Summary</th></tr>
<tr><th>Distribution</th><th style="width: 10em;"><table cellpadding="0" cellspacing="0" border="0"><tr><td align="left" style="border: none;"><img src="images/environment-spacer.gif" width="17" height="17" alt="" /></td><td width="100%" style="border: none;">Status</td><td align="right" style="border: none;"><img src="images/status_expand.png" width="16" height="16" alt="[&gt;&gt;]" id="expand-img" style="cursor: pointer;" onclick="javascript:toggle_table();" /></td></tr></table></th></tr>
DATA

	my %status_sort_order = ( 'OK' => 0, 'UNNEEDED' => 1, 'UNKNOWN' => 2, 'MISSED' => 3 );
	
	local *status_sort = sub {
		my ($a, $b) = @_;
		
		my $status_a = $distr_status{$a};
		my $status_b = $distr_status{$b};
		
		return (   $status_sort_order{$status_a} <=> $status_sort_order{$status_b}
				or ($lib_statistics{$a}{$status_a} or 0) <=> ($lib_statistics{$b}{$status_b} or 0)
				or ($int_statistics{$a}{$status_a} or 0) <=> ($int_statistics{$b}{$status_b} or 0)
				or $distr_names{$a} cmp $distr_names{$b} 
			);
	};
	
	# Distributions list
	foreach my $Did ( sort {status_sort($a, $b)} keys %$all_distrs ) {
		my $status = $distr_status{$Did};

		my $bgcolor;
		my $classname;
		if ( $status eq 'MISSED' ) {
			$bgcolor = "#ffdddd"; # red
			$classname = 'fail-td';
			$status = 'FAILED';
		}
		elsif ( $status eq 'UNKNOWN' || $status eq 'UNNEEDED' ) {
			$bgcolor = "#fefecc"; # yellow
			$classname = 'warn-td';
			$status = 'WARNINGS';
		}
		else {
			$bgcolor = "#ccffcc"; # green
			$classname = 'pass-td';
			$status = 'OK';
		}
		
		my $missed_libs   = ($lib_statistics{$Did}{MISSED} or 0);
		my $unneeded_libs = ($lib_statistics{$Did}{UNNEEDED} or 0);
		my $unknown_libs  = ($lib_statistics{$Did}{UNKNOWN} or 0);
		
		my $missed_ints  = ($int_statistics{$Did}{MISSED} or 0);
		my $unknown_ints = ($int_statistics{$Did}{UNKNOWN} or 0);

		# Distribution
		my $Dname = $distr_names{$Did};
		my $Dname_link = $Dname;
		$Dname_link =~ s/^(.*)? \d.*$/$1/;
		$Dname_link =~ s/ /%20/g;
		$general_table .= <<DATA;
<tr id="dtr_$Did" style="background: $bgcolor;">
<td style="white-space: nowrap;"><b><a href="http://www.linuxfoundation.org/navigator/browse/distr.php?cmd=list-byname&amp;Dname=$Dname_link" target="_blank">$Dname</a></b></td>
<td align="center" class="$classname" onmouseover="javascript:show_info(this, $missed_libs, $unneeded_libs, $missed_ints, $unknown_libs, $unknown_ints);" onmouseout="javascript:hide_info();" onclick="javascript:toggle_table();"><b>$status</b></td>
</tr>
DATA

		$ext_table .= '<tr style="background: '.$bgcolor.';">';

		# Missed libraries
		if ($missed_libs + $unneeded_libs == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			my $unneeded_libs_txt = ($unneeded_libs ? ' (+<font color="red"><b>'.$unneeded_libs.'</b></font> unneeded)' : '');
			my $missed_libs_txt = ($missed_libs ? '<font color="red"><b>'.$missed_libs.'</b></font>' : '0');
			$ext_table .= '<td style="white-space: nowrap; border-left: none;">'.$missed_libs_txt.$unneeded_libs_txt.' (<a href="" onclick="return det_open('.$Did.', \'missed_libs\')">list&hellip;</a>)</td>';
		}

		# Unknown libraries
		if ($unknown_libs == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			$ext_table .= '<td style="white-space: nowrap; border-left: none;"><font color="red"><b>'.$unknown_libs.'</b></font> (<a href="" onclick="return det_open('.$Did.', \'unknown_libs\')">list&hellip;</a>)</td>';
		}

		# Missed interfaces
		if ($missed_ints == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			$ext_table .= '<td style="white-space: nowrap; border-left: none;"><font color="red"><b>'.$missed_ints.'</b></font> (<a href="" onclick="return det_open('.$Did.', \'missed_ints\')">list&hellip;</a>)</td>';
		}

		# Unknown interfaces
		if ($unknown_ints == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			$ext_table .= '<td style="white-space: nowrap; border-left: none;"><font color="red"><b>'.$unknown_ints.'</b></font> (<a href="" onclick="return det_open('.$Did.', \'unknown_ints\')">list&hellip;</a>)</td>';
		}

		$ext_table .= '</tr>'.$nl;
	}
	$general_table .= '</table>';
	$ext_table .= '</table>';

	$res .= <<HEREDOC;
<table cellpadding="0" cellspacing="0" border="0" style="margin: 12px 0;">
<tr>
<td>
$general_table
</td>
<td>
$ext_table
</td>
</tr>
</table>

<script type="text/javascript" language="javascript">
// <![CDATA[

// Determines absolute dynamic position of an element (right top corner).
function getPosition(elem) {
	if (!elem)
		return {x: -1, y: -1};
	var x = elem.offsetWidth;
	var y = 0;
	while (elem) {
		x += elem.offsetLeft;
		y += elem.offsetTop;
		elem = elem.offsetParent;
	}
	return {x: x, y: y};
}

var distro_details = document.getElementById('distro-details');
var table = document.getElementById('d_table');
var table_ext = document.getElementById('d_table_ext');
var expand_img = document.getElementById('expand-img');

// Fine-tuning for browsers
var dDetailsOffsetX = ((navigator.appName == 'Netscape') ? 0 : 1);
if (navigator.appName == 'Konqueror') {
	table_ext.style.borderLeft = 'solid 1px #aaaaaa';
}

function show_info(td_elem, missed_libs, missed_zero_libs, missed_ints, unknown_libs, unknown_ints) {
	if (table_ext.style.display != 'none')
		return;
	var details = '';

	if (missed_libs + missed_zero_libs + missed_ints > 0) {
		var tmp = ''
		if (missed_libs + missed_zero_libs > 0) {
			tmp += (missed_libs ? ('<font color="red"><b>' + missed_libs + '</b></font> ' + ((missed_libs > 1) ? 'libraries' : 'library')) : '0 libraries') +
			       (missed_zero_libs ? ' (+<font color="red"><b>' + missed_zero_libs + '</b></font> unneeded)' : '');
		}
		if (missed_ints > 0) {
			tmp += (tmp == '' ? '' : '; ') +
			       (missed_ints ? ('<font color="red"><b>' + missed_ints + '</b></font> ' + ((missed_ints > 1) ? 'interfaces' : 'interface')) : '0 interfaces');
		}
		details += 'Missing: ' + tmp + '.';
	}
	if (unknown_libs + unknown_ints > 0) {
		var tmp = ''
		if (unknown_libs > 0) {
			tmp += (unknown_libs ? ('<font color="red"><b>' + unknown_libs + '</b></font> ' + ((unknown_libs > 1) ? 'libraries' : 'library')) : '0 libraries');
		}
		if (unknown_ints > 0) {
			tmp += (tmp == '' ? '' : '; ') +
			       (unknown_ints ? ('<font color="red"><b>' + unknown_ints + '</b></font> ' + ((unknown_ints > 1) ? 'interfaces' : 'interface')) : '0 interfaces');
		}
		details += (details == '' ? '' : ' ') + 'Unknown: ' + tmp + '.';
	}
	if (details == '')
		details = '<i>No problems detected.</i>';
	distro_details.innerHTML = details;
	var pos = getPosition(td_elem);
	distro_details.style.left = (pos.x + dDetailsOffsetX) + 'px';
	distro_details.style.top = pos.y + 'px';
	distro_details.style.height = (td_elem.offsetHeight - 1) + 'px';
	distro_details.style.display = '';
}

function hide_info() {
	distro_details.style.display = 'none';
}

function toggle_table() {
	hide_info();
	if (table_ext.style.display == 'none') {
		expand_img.setAttribute('src', 'images/status_collapse.png');
		expand_img.setAttribute('alt', '[&lt;&lt;]');
		table_ext.style.display = '';
	}
	else {
		expand_img.setAttribute('src', 'images/status_expand.png');
		expand_img.setAttribute('alt', '[&gt;&gt;]');
		table_ext.style.display = 'none';
	}
}
// ]]>
</script>
HEREDOC

	return $res;
}

# Contents of a detail popup window for Distributions table
sub distr_details {
	my ($Did, $show) = @_;

	my $nl = "\n";
	my $res = "";

	my $distr = $Analysis::all_distrs->{$Did};

	if ( !$distr ) {
		$res .= "There is no distribution with ID '$Did'.";
		return $res;
	}

	my %missed_libs = ();
	my %unknown_libs = ();
	my %unneeded_libs = ();
	my %present_libs = ();
	foreach my $lib ( values %$Analysis::libs ) {
		my $soname = $lib->{SONAME};
		if ( $lib->{DISTRS}{$Did} ) {
			$present_libs{$soname} = 1;
		} else {
			if ( !$lib->{APPROVED} ) {
				if ( ! $lib->{INTS_USED} ) {
					$unknown_libs{$soname} = 0;
				} else {
					$unknown_libs{$soname} = 1;
				}
			}
			else {
				if ( ! $lib->{INTS_USED} ) {
					$missed_libs{$soname} = 0;
				} else {
					$missed_libs{$soname} = 1;
				}
			}
		}
		if ( ! $lib->{INTS_USED} ) {
			$unneeded_libs{$soname} = 1;
		}
	}
	
	my %missed_ints = ();
	my %unknown_ints = ();
	#my %present_ints = ();
	foreach my $Iname ( keys %$Analysis::ints ) {
		foreach my $ver ( keys %{$Analysis::ints->{$Iname}} ) {
			my $int_obj = $Analysis::ints->{$Iname}{$ver};
			my $disp_name = ($int_obj->{UNMANGLED} or $Iname).( $ver ? " ".$ver : '' );
			if ( !$int_obj->{DISTRS}{$Did} ) {
				if ( $int_obj->{UNAPPROVED} ) {
					$unknown_ints{$disp_name} = $int_obj;
				} else {
					$missed_ints{$disp_name} = $int_obj;
				}
			}
			#else {
				#$present_ints{$disp_name} = $int_obj;
			#}
		}
	}

	my $Dname = distr_name($distr);
	my $Dname_link = $Dname;
	$Dname_link =~ s/^(.*)? \d.*$/$1/;
	$Dname_link =~ s/ /%20/g;
	$res .= '<h2>Distribution: <a href="http://www.linuxfoundation.org/navigator/browse/distr.php?cmd=list-byname&amp;Dname='.$Dname_link.'" target="_blank"><u>'.$Dname.'</u></a></h2>'.$nl;

	local *lib_list = sub {
		my ($hash, $color, $color2) = @_;

		my $res = "";

		$res .= '<ul>'.$nl;
		foreach my $soname ( sort {lc $a cmp lc $b} keys %$hash ) {
			my $link = href( tab_link('libs').'&lib='.$soname );
			$res .= '<li>';
			my $href_color = $color;
			$href_color = ($color2 or '#907000') if $hash->{$soname} == 0;
			$res .= '<a style="color:'.$href_color.'" href="'.html($link).'">'.html($soname).'</a>';
			$res .= ' &nbsp; (<i>0 interfaces used</i>)' if $hash->{$soname} == 0;
			$res .= '</li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};

	if ( %missed_libs && (!$show || $show eq 'missed_libs') ) {
		$res .= '<h3>Missing Libraries:</h3>'.$nl;

		$res .= lib_list(\%missed_libs, 'red');
	}

	if ( %unknown_libs && (!$show || $show eq 'unknown_libs') ) {
		$res .= '<h3>No Information About:</h3>'.$nl;
		$res .= '<p>The Application Checker Database contains no information'
			.' about the following libraries.</p>'.$nl;
		
		$res .= lib_list(\%unknown_libs, '#907000'); # yellow
	}

	if ( %unneeded_libs && (!$show || $show eq 'unneeded_libs') ) {
		$res .= '<h3>Unnecessary Libraries:</h3>'.$nl;
		$res .= '<p>These are libraries which are required by the Application,'
			.' but <b>no interfaces</b> from these libraries are actually used.</p>'.$nl;

		$res .= lib_list(\%unneeded_libs, '#907000'); # yellow
	}
	
	if ( %present_libs && (!$show || $show eq 'present_libs') ) {
		$res .= '<h3>Present Libraries:</h3>'.$nl;

		$res .= lib_list(\%present_libs, 'green');
	}

	$res .= '<hr/>' if !$show;
	
	local *int_list = sub {
		my ($color, $hash) = @_;

		my $res = "";

		$res .= '<ul>'.$nl;
		foreach my $disp_name ( sort {lc $a cmp lc $b} keys %$hash ) {
			my $int_obj = $hash->{$disp_name};
			my $fullname = $int_obj->{NAME}.( $int_obj->{VER} ? ' '.$int_obj->{VER} : '');
			my $libraries = join ",", @{$int_obj->{LIBRARIES}};
			my $link = href( tab_link('ints').'&int='.$fullname );
			$res .= '<li>';
			$res .= '<a style="color:'.$color.'" href="'.html($link).'">'.html($disp_name).'</a>';
			$res .= ' &nbsp; ('.html($libraries).')' if $libraries;
			$res .= '</li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};

	if ( %missed_ints && (!$show || $show eq 'missed_ints') ) {
		$res .= '<h3>Missing Interfaces:</h3>'.$nl;

		$res .= int_list('red', \%missed_ints);
	}
	
	if ( %unknown_ints && (!$show || $show eq 'unknown_ints') ) {
		$res .= '<h3>No Information About:</h3>'.$nl;
		$res .= '<p>The Application Checker Database contains no information'
			.' about the following interfaces.</p>'.$nl;
		
		$res .= int_list('#907000', \%unknown_ints); # yellow
	}

	#if ( %present_ints ) {
		#$res .= '<h3>Present Interfaces:</h3>'.$nl;

		#$res .= lib_list('green', \%present_ints);
	#}
	
	$res .= '<hr/>' if !$show;
	
	if ( $show ) {
		$res .= '<hr/>';
		my $link = href( tab_link('distrs').'&id='.$Did );
		$res .= '<p><a href="'.html($link).'"><i>More details&hellip;</i></a></p>';
	}
	
	return $res;
}

# Renders main title of the results page.
sub get_title {
	my ($app_name, $architecture, $res_dir) = @_;

	my $res = '';
	if (! -f "$res_dir/test_logs/ftp_upload") {
		$res .= <<DATA;
<table border="0" cellpadding="0" cellspacing="0" class="upload-suggest" onclick="javascript:document.location='upload_data.pl?details=$_GET{'details'}';">
  <tr>
    <td valign="middle"><img src="images/upload_data.png" width="16" height="16" alt="" /></td>
    <td><img src="images/environment-spacer.gif" width="3" height="1" alt="" /></td>
    <td>Please <a href="upload_data.pl?details=$_GET{'details'}" onclick="javascript:event.cancelBubble=true;"><b>upload</b></a> info about your application to the <a href="http://www.linuxfoundation.org/navigator/browse/app.php" target="_blank" onclick="javascript:event.cancelBubble=true;">LSB Navigator</a>.</td>
  </tr>
</table>
DATA
	$res .= "<h1>Analysis Results for <font color=\"blue\">$app_name</font> on <font color=\"blue\">$architecture</font></h1>\n\n";

	}
	return $res;
}

# Renders a short summary for the header of the results page.
sub summary {
	my ($selected_lsb_ver) = @_;

	## Calc Distrs status
	
	my $total_distrs = scalar keys %$Analysis::all_distrs;
	
	my $unsupp_distrs = 0;
	my $warn_distrs = 0;
	my $good_distrs = 0;

	foreach my $Did ( keys %$Analysis::all_distrs ) {
		my ($status, $statistics) = Analysis::distr_status($Did); # OK|UNNEEDED|UNKNOWN|MISSED
		if ( $status eq 'OK' ) { $good_distrs++; }
		if ( $status eq 'UNNEEDED' || $status eq 'UNKNOWN' ) { $warn_distrs++; }
		if ( $status eq 'MISSED' ) { $unsupp_distrs++; }
	}
	
	## Calc LSB status for libraries
	
	my %lib_lsb_status = ();
	my %lib_lsb_status_ver = ();
	
	my $zero_int_libs = 0;
	my $unapproved_libs = 0;

	foreach my $lib ( values %$Analysis::libs ) {

		# Check libs without interfaces used
		if ( !$lib->{INTS_USED} ) {
			$zero_int_libs++;
		}
		if ( !$lib->{APPROVED} ) {
			$unapproved_libs++;
		}

		# Check LSB
		my ($status, $status_ver) = lsb_status( $lib, $selected_lsb_ver );
		$status_ver = '' if not defined $status_ver;
		
		$lib_lsb_status{$status}++;
		$lib_lsb_status_ver{$status}{$status_ver}++;
	}

	## Calc LSB status for interfaces
	
	my %int_lsb_status = ();
	my %int_lsb_status_ver = ();

	foreach my $int_vers ( values %$Analysis::ints ) {
		foreach my $int ( values %$int_vers ) {

			my ($status, $status_ver) = int_lsb_status( $int, $selected_lsb_ver );
			$status_ver = '' if not defined $status_ver;
			
			$int_lsb_status{$status}++;
			$int_lsb_status_ver{$status}{$status_ver}++;
		}
	}

	## Let's print the summary

	my $nl = "\n";
	my $res = '';

	my $green_image  = '<img src="images/dot_green.png" width="16" height="16" alt="green" /> ';
	my $yellow_image = '<img src="images/dot_yellow.png" width="16" height="16" alt="yellow" /> ';
	my $red_image    = '<img src="images/dot_red.png" width="16" height="16" alt="red" /> ';
	my $warn_image   = '<img src="images/dot_warn.png" width="16" height="16" alt="warning" /> ';

	my $green_table = <<DATA;
<div style="background-color: #40d040; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>No compatibility problems detected</b>
</div>
<div id="summary" style="border: 1px solid #7f7f7f; border-top: none; margin-top: 0; padding: 0.3em;">
DATA
	my $yellow_table = <<DATA;
<div style="background-color: #d0a720; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>Some compatibility problems detected</b>
</div>
<div id="summary" style="border: 1px solid #7f7f7f; border-top: none; margin-top: 0; padding: 0.3em;">
DATA
	my $red_table = <<DATA;
<div style="background-color: #d04040; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>Numerous compatibility problems found</b>
</div>
<div id="summary" style="border: 1px solid #7f7f7f; border-top: none; margin-top: 0; padding: 0.3em;">
DATA

	local *_pl = sub {
		my ($num, $mid, $singular, $plural) = @_;
		
		return '<b>'.$num.'</b> '.($mid ? $mid.' ' : '').($num > 1 ? $plural : $singular);
	};
	
	local *refer = sub {
		my ($text, $tab, $add, $no_below) = @_;
		
		if ( $_GET{"tab"} ne $tab ) {
			my $link = href(tab_link($tab));
			if ( $add ) { $link .= href($add); }
			return '<a href="'.html($link).'">'.$text.'</a>';
		} else {
			$text .= ' (see below)' unless $no_below;
			return $text;
		}
	};
	
	local *see_details = sub {
		my ($tab, $add) = @_;
		
		if ( $_GET{"tab"} ne $tab ) {
			my $link = href(tab_link($tab));
			if ( $add ) { $link .= href($add); }
			return ' (<a href="'.html($link).'">details&hellip;</a>)';
		} else {
			return ' (see the table below)';
		}
	};

	# Summary on compatibility with various distributions
	if ( $warn_distrs == 0 && $unsupp_distrs == 0 ) {
		$res .= $green_table;
		if ( %$Analysis::libs ) {
			$res .= '<p>';
			$res .= $green_image; # Green image
			$res .= 'All libraries and interfaces used by the Application';
			$res .= ' are present in all the analyzed distributions';
			$res .= see_details('distrs', '&expand=1');
			$res .= '.';
		}
	}
	elsif ( $warn_distrs > 0 && $unsupp_distrs == 0 ) {
		$res .= $green_table;
		$res .= '<p>';
		$res .= $green_image; # Green image
		$res .= 'All libraries and interfaces required by the Application'
				.' (except some unused or unknown libraries)'
				.' are present in all the analyzed distributions';
		$res .= see_details('distrs', '&expand=1');
		$res .= '.';
	}
	elsif ( $good_distrs > 0 && $warn_distrs == 0 && $unsupp_distrs > 0 ) {
		$res .= $yellow_table;
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		$res .= 'There '.($good_distrs > 1 ? 'are' : 'is' ).' '
				.refer('<b>'.$good_distrs.'</b> of <b>'.$total_distrs.'</b> distributions', 'distrs', '&expand=1')
				.' that provide all the required libraries and interfaces.';
	}
	elsif ( $warn_distrs > 0 && $unsupp_distrs > 0 ) {
		$res .= $yellow_table;
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		my $n = $good_distrs + $warn_distrs;
		$res .= 'There '.($n > 1 ? 'are' : 'is' ).' '
				.refer('<b>'.$n.'</b> of <b>'.$total_distrs.'</b> distributions', 'distrs', '&expand=1')
				.' that provide all the required libraries and interfaces'
				.' except unneeded and unknown.';
	}
	elsif ( $good_distrs == 0 && $warn_distrs == 0 && $unsupp_distrs > 0 ) {
		$res .= $red_table;
		$res .= '<p>';
		$res .= $red_image; # Red image
		$res .= 'The Application is unlikely to run on any of the analyzed distributions.';
	}
	$res .= '</p>'.$nl;

	# Summary on LSB compatibility
	$res .= '<p>';
	if ( $lib_lsb_status{"NOT"} || $lib_lsb_status{"WITHDRAWN"} ) {
		my $x = ($lib_lsb_status{"NOT"} or 0) + ($lib_lsb_status{"WITHDRAWN"} or 0)
				+($lib_lsb_status{"CANDIDATE"} or 0) + ($lib_lsb_status{"FUTURE"} or 0);
		$res .= $red_image; # Red image
		$res .= 'The Application uses '
			.refer( _pl($x, 'external', 'library', 'libraries'), 'libs', '&show=non-lsb' )
			.' <b>incompatible with LSB '.$selected_lsb_ver.'</b>.';
	}
	elsif ( $int_lsb_status{"NOT"} || $int_lsb_status{"WITHDRAWN"} ) {
		my $x = ($int_lsb_status{"NOT"} or 0) + ($int_lsb_status{"WITHDRAWN"} or 0)
				+($int_lsb_status{"CANDIDATE"} or 0) + ($int_lsb_status{"FUTURE"} or 0);
		$res .= $red_image; # Red image
		$res .= 'The Application uses '
			.refer( _pl($x, '', 'interface', 'interfaces'), 'ints', '&show=non-lsb')
			.' <b>incompatible with LSB '.$selected_lsb_ver.'</b>.';
	}
	elsif ( $lib_lsb_status{"CANDIDATE"} ) {
		my $cand_version = (reverse sort keys %{$lib_lsb_status_ver{"CANDIDATE"}})[0]; # max

		$res .= $yellow_image; # Yellow image
				
		if ( $lib_lsb_status{"FUTURE"} ) {
			# TODO: vvv - text for mixed state CANDIDATE & FUTURE
			$res .= 'All libraries used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or future versions or are candidates'
				.' for LSB '.$cand_version.'.';
		} else {
			$res .= 'All libraries used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' version or are candidates'
				.' for LSB '.$cand_version.'.';
		}
		
		if ( $lib_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $int_lsb_status{"CANDIDATE"} ) {
		my $cand_version = (reverse sort keys %{$int_lsb_status_ver{"CANDIDATE"}})[0]; # max
		
		$res .= $yellow_image; # Yellow image
		
		if ( $int_lsb_status{"FUTURE"} ) {
			# TODO: vvv - text for mixed state CANDIDATE & FUTURE
			$res .= 'All interfaces used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or future versions or are candidates'
				.' for LSB '.$cand_version.'.';
		} else {
		
			$res .= 'All interfaces used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' version or are candidates'
				.' for LSB '.$cand_version.'.';
		}
		
		if ( $int_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some interfaces used by the Application are <b>deprecated</b>.';
		}
		elsif ( $lib_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $lib_lsb_status{"FUTURE"} ) {
		my $future_version = (reverse sort keys %{$lib_lsb_status_ver{"FUTURE"}})[0]; # max
		
		my $x = $lib_lsb_status{"FUTURE"};
		
		$res .= $yellow_image; # Yellow image
		$res .= 'The Application uses '._pl($x, '', 'library', 'libraries')
				.' which appeared in a <b>later version of the LSB</b>.';
		$res .= ' Try checking the Application for compatibility with <b>LSB '.$future_version.'</b>.';
		
		if ( $lib_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
	}
	elsif ( $int_lsb_status{"FUTURE"} ) {
		my $future_version = (reverse sort keys %{$int_lsb_status_ver{"FUTURE"}})[0]; # max
		
		my $x = $int_lsb_status{"FUTURE"};
		
		$res .= $yellow_image; # Yellow image
		$res .= 'The Application uses '._pl($x, '', 'interface', 'interfaces')
				.' which appeared in <b>later version of the LSB</b>.';
		$res .= ' Try checking it for compatibility with <b>LSB '.$future_version.'</b>.';
		
		if ( $int_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some interfaces used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
		elsif ( $lib_lsb_status{"DEPRECATED"} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
	}
	elsif ( $lib_lsb_status{"DEPRECATED"} || $int_lsb_status{"DEPRECATED"} ) {
		$res .= $yellow_image; # Yellow image
		$res .= 'All libraries and interfaces used by the Application';
		$res .= ' are <b>included in LSB '.$selected_lsb_ver.'</b>';
		$res .= ', but some of them are <b>deprecated</b>.';
		$res .= ' You may have compatibility issues with new distributions in the future.';		
	}
	else { #  "IN" or "OPTIONAL"
		$res .= $green_image; # Green image
		$res .= 'All libraries and interfaces used by the Application';
		$res .= ' are <b>included in LSB '.$selected_lsb_ver.'</b>.';
		$res .= ' You have a good chance to achieve '.refer('LSB compliance', 'cert', '', 1).'.';
	}
	$res .= '</p>'.$nl;

	# Libraries without interfaces used
	if ( $zero_int_libs ) {
		$res .= '<p>';
		$res .= $warn_image; # Yellow image
		$res .= 'The Application uses '
			.refer( _pl($zero_int_libs, 'external', 'library', 'libraries'), 'libs', '&show=unneeded' )
			.' <b>without using interfaces</b> in '.(($zero_int_libs > 1) ? 'those libraries' : 'this library').'.'
			.' The Application would be more portable if it is fixed to not use unneeded libraries.';
		$res .= '</p>';
	}
	
	# Unapproved libraries
	if ( $unapproved_libs ) {
		$res .= '<p>';
		$res .= $warn_image; # Yellow image
		$res .= 'The Application uses '
			.refer( _pl($unapproved_libs, 'external', 'library', 'libraries'), 'libs', '&show=unapproved' )
			.' that are <b>not common</b> in distributions.'
			.' Please consider providing such libraries as a part of the Application package.';
		$res .= '</p>';
	}

	$res .= "$nl</div>$nl<br />$nl";
	$res .= <<DATA;

<div id="tooltip" style="display: none; background-color: #ffffd0; border: solid 1px black; position: absolute; z-index: 1000;">
<table cellpadding="0" cellspacing="0" border="0" style="width: 17em;">
  <tr style="background-color: #d0d0d0; color: black;">
    <td align="right" style="border-bottom: solid 1px black;"><img src="images/environment-spacer.gif" width="7" height="1" alt="" /></td>
    <td align="center" width="100%" style="border-bottom: solid 1px black;"><b>Information</b></td>
    <td align="right" style="border-bottom: solid 1px black;"><img src="images/close.png" onclick="javascript:initDialog(null);" width="20" height="20" title="Close" alt="X" style="cursor: pointer;" /></td>
  </tr>
  <tr>
    <td colspan="3" style="padding: 5px; white-space: nowrap;" id="tooltip_text"></td>
  </tr>
</table>
</div>

<table cellpadding="0" cellspacing="0" border="0" style="display: none; position: absolute; z-index: 999;" id="shadow">
  <tr>
    <td width="14" height="14" style="background: url('images/shadow_tl.png') no-repeat;"></td>
    <td height="14" style="background: url('images/shadow_tc.png') repeat-x;"></td>
    <td width="14" height="14" style="background: url('images/shadow_tr.png') no-repeat;"></td>
  </tr>
  <tr>
    <td width="14" style="background: url('images/shadow_ml.png') repeat-y;"></td>
    <td style="background: url('images/shadow_mc.png') repeat;" id="shadow_center"></td>
    <td width="14" style="background: url('images/shadow_mr.png') repeat-y;"></td>
  </tr>
  <tr>
    <td width="14" height="14" style="background: url('images/shadow_bl.png') no-repeat;"></td>
    <td height="14" style="background: url('images/shadow_bc.png') repeat-x;"></td>
    <td width="14" height="14" style="background: url('images/shadow_br.png') no-repeat;"></td>
  </tr>
</table>

<script type="text/javascript" language="javascript">
// <![CDATA[

preload_images(new Array('status_collapse.png', 'close.png', 'shadow_bc.png', 'shadow_bl.png', 'shadow_br.png', 'shadow_mc.png', 'shadow_ml.png', 'shadow_mr.png', 'shadow_tc.png', 'shadow_tl.png', 'shadow_tr.png'));

var just_shown_dialog = false;          // Indicates that the dialog was just shown (to prevent its hiding from onBodyClick)
var dialog = null;                      // The dialog itself
var dialog_caller = null;               // Element which caused the dialog to appear
var dialog_caller_class = null;         // Normal class name of the dialog caller (to be restored when the dialog is closed)
var dialog_shadow = document.getElementById('shadow');
var dialog_shadow_center = document.getElementById('shadow_center');

// Hides the previous dialog (if it was shown) and prepare data for
// displaying the new one.
// CALLER - the HTML element which caused the dialog to appear and (possibly)
// will change its appearance because of that.
function initDialog(caller) {
	if (dialog_caller && dialog_caller_class)
		dialog_caller.className = dialog_caller_class;
	dialog_shadow.style.display = 'none';
	if (dialog)
		dialog.style.display = 'none';
	if (!caller || (caller == dialog_caller)) {
		dialog_caller = null;
		dialog_caller_class = null;
		return false;
	}
	else {
		dialog_caller = caller;
		dialog_caller_class = caller.className;
		return true;
	}
}

// Shows the dialog at the specified position and adjusts the shadow below it.
function showDialog(show_dlg, pos) {
	dialog = show_dlg;
	dialog.style.left = pos.x + 'px';
	dialog.style.top = pos.y + 'px';
	dialog.style.display = '';
	just_shown_dialog = true;
	dialog_shadow_center.width = dialog.offsetWidth - 15;
	dialog_shadow_center.height = dialog.offsetHeight - 15;
	dialog_shadow.style.left = pos.x + 'px';
	dialog_shadow.style.top = pos.y + 'px';
	dialog_shadow.style.display = '';
}

function onBodyClick() {
	if (!just_shown_dialog)
		initDialog(null);
	just_shown_dialog = false;
}

// Tooltip processing functions
var tooltip_texts = new Array();
tooltip_texts['libs_ints_common'] = 'Common libraries are manually selected by the Linux Foundation<br />to represent .so names of those libraries which may be externally required<br />by applications and provided by distributions.<br />Please write to <a href="mailto:lsb-appcheck-support\@linuxfoundation.org">lsb-appcheck-support\@linuxfoundation.org</a> to suggest adding more libraries<br />to the "common libraries" list.';
tooltip_texts['libs_ints_unknown'] = 'Unknown libraries are those not included in the "approved" list<br />of common libraries. This means such libraries are not supposed<br />to be externally required by applications and correspondingly provided<br />by distributions. Sometimes new versions of common libraries<br />fall into the "unknown" category as well. It is recommended to include<br />such libraries as a part of your application (either by statically linking<br />or by supplying the .so in your package).';
tooltip_texts['ints_linkage_failed'] = 'The analyzer has failed to determine which library (among those required<br />by this application component) provides this interface.<br />This may happen if the corresponding library wasn\\'t found<br />in the database or in the distribution on this machine, or<br />if the list of libraries needed by this application component<br />is incomplete. If this binary is executable, then most likely it<br />will fail if launched on this machine.';

function show_hide_tooltip(elem_parent, id, event) {
	if (!initDialog(elem_parent))
		return;
	var tt_window = document.getElementById('tooltip');
	if (tt_window) {
		var tt_text = document.getElementById('tooltip_text');
		if (tt_text) {
			if (tooltip_texts[id])
				tt_text.innerHTML = tooltip_texts[id];
			else
				tt_text.innerHTML = '<font color="red"><b>Sorry, no information for this item was found!</b></font><br />Please write to <a href="mailto:lsb-appcheck-support\@linuxfoundation.org">lsb-appcheck-support\@linuxfoundation.org</a>.';
			var scrollTop = document.body.scrollTop;
			if (scrollTop == 0)
			{
			    if (window.pageYOffset)
			        scrollTop = window.pageYOffset;
			    else
			        scrollTop = (document.body.parentElement) ? document.body.parentElement.scrollTop : 0;
			}
			var scrollLeft = document.body.scrollLeft;
			if (scrollLeft == 0)
			{
			    if (window.pageXOffset)
			        scrollLeft = window.pageXOffset;
			    else
			        scrollLeft = (document.body.parentElement) ? document.body.parentElement.scrollLeft : 0;
			}
			var div_top = event.clientY + scrollTop + 10;
			var div_left = event.clientX + scrollLeft + 10;
			tt_window.style.display = '';
			var div_width = (tt_window.style.pixelWidth || tt_window.offsetWidth);
			var wnd_width = (document.body.clientWidth || window.innerWidth);
			if (div_left + div_width > wnd_width + scrollLeft)
				div_left = wnd_width + scrollLeft - div_width;
			if (div_left < scrollLeft)
				div_left = scrollLeft;
			showDialog(tt_window, {x: div_left, y: div_top});
		}
	}
}

// ]]>
</script>
DATA

	return $res;
}

# Library compatibility table
sub I_table {
	my ($selected_lsb_ver) = @_;

	my $all_distrs = $Analysis::all_distrs;
	
	my $res = "";

	my $nl = "\n"; # Set this to "" to make spaghetti HTML code

	if ( !%$Analysis::ints ) {
		$res .= no_binary_div();
		return $res;
	}

	$res .= <<HEREDOC;
The table below lists all external interfaces required by your Application (based on ELF symbols). LSB and Distribution presence statuses are provided for each interface.<br /><br />
HEREDOC

	my $link = href( tab_link('ints') );
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id )
{
	var v = document.getElementById( id );
	if ( v ) {
		v.style.display = (v.style.display == 'none' ? '' : 'none');
	}
	return false;
}

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes'
		);
	w.moveTo(100,0);
	w.focus();

	return w;
}

function det_open( int )
{
	popup_win( '$link' + '&int=' + int );

	return false;
}

// ]]>
</script>
HEREDOC
	
	my $by_name = {};
	foreach my $Iname ( keys %$Analysis::ints ) {
		foreach my $int_obj ( values %{$Analysis::ints->{$Iname}} ) {
			my $name = ($int_obj->{UNMANGLED} or $int_obj->{NAME});
			$name .= '@'.$int_obj->{VER} if $int_obj->{VER};
			
			$by_name->{$name} = $int_obj;
		}
	}
	
	my $table = "";
	# Table Header
	my $table_open = '<table id="i_table" class="wikitable" width="100%">';
	$table .= $table_open.$nl;
	$table .= '<tr>'
		.'<th>Interface</th>'
		.'<th>Version</th>'
		.'<th>Dependent<br/>Application&nbsp;Components</th>'
		.'<th>In&nbsp;Libraries</th>'
		.'<th>Presence in<br />Distributions</th>'
		.'<th id="lsb_col_header" style="white-space:nowrap">In&nbsp;LSB&nbsp;'.$selected_lsb_ver.'?</th>'
		.'<th>More Info</th>'
		.'</tr>';

	my $filt_show = {};
	my $filt_library = {};
	my $filt_component = {};

	my $row_n = 0; # Counter for generating row-specific IDs
	
	# Table Body
	foreach my $name ( sort {lc $a cmp lc $b} keys %$by_name ) {
		my $int_obj = $by_name->{$name};
		my $Iname = $int_obj->{NAME};
		
		my $row = "";
		$row_n++;

		$row .= '<tr>'.$nl;

		## Interface
		my $unmangled = ($int_obj->{UNMANGLED} or $Iname);
		
		# wrap unmangled name (by _) if it is too long
		my $tmp = "";
		$unmangled =~ s/\(/ \(/g;
		$unmangled =~ s/\)/\) /g;
		while ( $unmangled =~ s/^(\S{0,40}?\s+|\S{30}[^_]{0,20}|\S{1,40}$)// ) {
			$tmp .= html($&);
			$tmp .= "<br/>" if $unmangled && $& !~ /\s$/;
		}
		$unmangled = $tmp;
		
		my $iname_link = href( 'http://www.linuxfoundation.org/navigator/browse/int_single.php'
			.'?cmd=list-by-name&Iname='.$Iname );
		if ( $int_obj->{LIBRARIES} ) {
			my ($library) = @{$int_obj->{LIBRARIES}};
			$library =~ s/\.so.*//; # get the library name
			$library =~ s/libstdc\+\+/libstdcxx/; # fix for the database
			$iname_link .= href('&Ilibrary='.$library);
		}

		$row .= '<td><a href="'.html($iname_link).'" target="_blank">'.$unmangled.'</a></td>'.$nl;
		
		# Version
		$row .= '<td>'.($int_obj->{VER} or "").'</td>'.$nl;
		
		# App Modules: preparation
		my $app_modules = "";
		my $app_comp_n = 0;
		if ( $int_obj->{APP_MODULES} ) {
			foreach my $component ( sort @{$int_obj->{APP_MODULES}} ) {
				my ($disp_name, $disp_path) = parse_app_component($component);
				
				$app_modules .= (($app_modules eq '') ? '' : '<br />');
				$app_modules .= '<span title="'.html_arg($disp_path).'">'.html($disp_name).'</span>';
				
				$app_comp_n++;
				
				$filt_component->{$disp_path}{$row_n} = 1;
			}
		}

		## App modules expandable list
		$row .= '<td align="left" style="white-space:nowrap">';
		my $div_id = "cl".$row_n;
		my $plu = ( $app_comp_n > 1 ? 's' : '' );
		$row .= '<a href="" onclick="return toggle(\''.$div_id.'\')">'
					# Required by
					.$app_comp_n.' component'.$plu
					.' (&hellip;)'
				.'</a>';
		$row .= '<div id="'.$div_id.'" style="display:none;text-align:left">';
		$row .= $app_modules;
		$row .= '</div>';
		$row .= '</td>'.$nl;

		# In Libraries
		my $libraries = "";
		foreach my $soname ( @{$int_obj->{LIBRARIES}} ) {
			$libraries .= ",<br />" if $libraries;
			$filt_library->{$soname}{$row_n} = 1;
			$soname_link = href( 'http://www.linuxfoundation.org/navigator/browse/rawlib.php'
				.'?cmd=list-byrunname&RLrunname='.$soname);
			if (   $Analysis::libs->{$soname} && !$Analysis::libs->{$soname}{APPROVED} 
				or $Analysis::libs_impl->{$soname} && !$Analysis::libs_impl->{$soname}{APPROVED} )
			{
				$soname = '<a href="'.html($soname_link).'" target="_blank" style="color:#907000">'.html($soname).'</a>';
			}
			else
			{
				$soname = '<a href="'.html($soname_link).'" target="_blank">'.html($soname).'</a>';
			}
			$libraries .= $soname;
		}
		$row .= '<td>'.$libraries.'</td>'.$nl;

		## Presence in distributions

		my $total = scalar( keys %$all_distrs );
		my $present = 0;
		if ( $int_obj->{DISTRS} ) {
			$present = scalar( keys %{$int_obj->{DISTRS}} );
		}

		my $bgcolor;
		if ( $present == $total ) {
			$bgcolor = "#CCFFCC"; # green
		} else {
			if ( $present > 0 ) {
				$bgcolor = "#FEFECC"; # yellow
			} else {
				if ( $int_obj->{UNAPPROVED} ) {
					$bgcolor = "#FEECCC"; # dark yellow
				} else {
					$bgcolor = "#FFCCCC"; # red
				}
			}
			$filt_show->{'problems'}{$row_n} = 1;
		}

		$row .= '<td align="right" style="background:'.$bgcolor.';white-space:nowrap">';
		if ( $present > 0 ) {
			$row .= $present." of ".$total;
			my $int = $int_obj->{NAME}.( $int_obj->{VER} ? '+'.$int_obj->{VER} : '' );
			my $info_link = href( tab_link('ints').'&int='.$int );
			my $info_onclick = 'onclick="return det_open(\''.html_arg($int).'\')"';
			$row .= ' (<a href="'.html($info_link).'" '.$info_onclick.'>list&hellip;</a>)';
		}
		elsif ( $int_obj->{OTHER_VER_FOUND} ) {
			$row .= "Unknown version";
		}
		else {
			if ( $int_obj->{UNAPPROVED} ) {
				$row .= 'Unknown&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, \'libs_ints_unknown\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
			} else {
				if ( $libraries ) {
					$row .= "Missing";
				} else {
					$row .= "Linkage Failed";
					# Tooltip:
					$row .= '&nbsp;<a href="tests_help.pl#ints_linkage_failed" onclick="javascript:show_hide_tooltip(this, \'ints_linkage_failed\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
				}
			}
		}
		
		$row .= '</td>'.$nl;

		## LSB
		$row .= '<td align="center">';

		$row .= '<div>';

		my ($lsb_status, $lsb_status_ver) = int_lsb_status($int_obj, $selected_lsb_ver);
		$lsb_status_ver = '' if !defined $lsb_status_ver;
		
		my $color = undef;
		
		if ( $lsb_status eq "NOT" || $lsb_status eq "WITHDRAWN" ) {
			$row .= '<span style="color:red">No</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "CANDIDATE" ) {
			$row .= '<span style="color:#907000">Candidate for '.$lsb_status_ver.'</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "FUTURE" ) {
			$row .= '<span style="color:#907000">In LSB '.$lsb_status_ver.'</span>';
			$filt_show->{'non-lsb'}{$row_n} = 1;
		}
		elsif ( $lsb_status eq "OPTIONAL" ) {
			$row .= '<span style="color:blue">Optional</span>';
		}
		elsif ( $lsb_status eq "IN" ) {
			$row .= '<span style="color:green">Yes</span>';
		}
		elsif ( $lsb_status eq "DEPRECATED" ) {
			$row .= '<span style="color:blue">Deprecated</span>';
		}

		$row .= '</div>';

		$row .= '</td>'.$nl;


		## More Info

		$row .= '<td>';
		my $int = $int_obj->{NAME}.( $int_obj->{VER} ? '+'.$int_obj->{VER} : '' );
		my $info_link = href( tab_link('ints').'&int='.$int );
		my $info_onclick = 'onclick="return det_open(\''.html($int).'\')"';
		my $txt = 'Details&hellip;';
		if ( $int_obj->{ATTR} ) {
			$txt = '<b>More&nbsp;Info&hellip;</b>';
		}
		# TODO: || $Analysis::intKBDistr->{$soname}
		$row .= '<a href="'.html($info_link).'" '.$info_onclick.'>'.$txt.'</a>';
		$row .= '</td>'.$nl;

		$row .= '</tr>'.$nl;
		
		$table .= $row;
	}
	
	# Table Footer
	$table .= '</table>'.$nl;
	
	my $interfaces_num = scalar keys %$by_name;
	
	# Filters before the table
	$res .= '<table><tr>';
	
	# "Component" filter
	$res .= '<td><table class="bartable">'
  		.'<tr>'
  		.'<td align="right"><b>Component:</b></td>'
    	.'<td align="right">'
      	.'<select name="com_select" id="com_select" onchange="return apply_filter();" style="width: 12.5em;">'
        .'<option value="" selected="selected">All</option>';
		foreach my $component ( sort {lc $a cmp lc $b} keys %$filt_component ) {
			$res .= '<option value="'.html_arg($component).'">'.html($component).'</option>';
		}
    $res .= '</select>'
    	.'</td>'
  		.'</tr>'
		.'</table></td>';
	
	# "Library" filter
	$res .= '<td><table class="bartable">'
  		.'<tr>'
  		.'<td align="right"><b>Library:</b></td>'
    	.'<td align="right">'
      	.'<select name="lib_select" id="lib_select" onchange="return apply_filter();" style="width: 12.5em;">'.$nl
        .'<option value="" selected="selected">All</option>'.$nl;
		foreach my $soname ( sort {lc $a cmp lc $b} keys %$filt_library ) {
			$res .= '<option value="'.html_arg($soname).'">'.html($soname).'</option>'.$nl;
		}
    $res .= '</select>'
    	.'</td>'
  		.'</tr>'
		.'</table></td>';
	
	# "Show" filter
	$res .= '<td><table class="bartable">'
  		.'<tr>'
  		.'<td align="right"><b>Show:</b></td>'
    	.'<td align="right">'
      	.'<select name="filt" id="filt" onchange="return apply_filter();" style="width: 12.5em;">'.$nl
        .'<option value="" selected="selected">All</option>'.$nl
        .'<option value="problems">Compatibility problems</option>'.$nl
        .'<option value="non-lsb">Non-LSB</option>'.$nl
      	.'</select>'
    	.'</td>'
  		.'</tr>'
		.'</table></td>';
		
	$res .= '<td id="lines_count" style="padding-left:20pt">'.$interfaces_num.' interface'.($interfaces_num!=1?'s':'').' shown.</td>';
	$res .= '<td>[<a href="" onclick="javascript:document.getElementById(\'filt\').value=\'\'; document.getElementById(\'lib_select\').value=\'\';document.getElementById(\'com_select\').value=\'\';return apply_filter();">Reset all filters</a>]</td>';
	$res .= '</tr></table>';
	
	## Here the Table goes
	$res .= '<div id="wrapper">'.$nl;
	$res .= $table;
	$res .= '</div>'.$nl;
	##
	
	# Javascript stuff
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
var wrapper = document.getElementById( 'wrapper' );
var table = document.getElementById( 'i_table' );
var lcp = document.getElementById( 'lines_count' );

var l = new Array();
var table_lines = table.getElementsByTagName('tr');
for ( var i=0; i<table_lines.length; i++ ) {
	l[i] = '<tr>' + table_lines[i].innerHTML + '</tr>';
}
HEREDOC

	# javascript: filt_problems
	$res .= 'var filt_show = Array();'.$nl;
	foreach my $what ( keys %$filt_show ) {
		$res .= 'filt_show[\''.$what.'\']=new Array(""';
		foreach my $row_n ( keys %{$filt_show->{$what}} ) {
			$res .= ','.$row_n;
		}
		$res .= ');'.$nl;
	}
	
	# javascript: filt_library
	$res .= 'var filt_lib = Array();'.$nl;
	foreach my $soname ( keys %$filt_library ) {
		$res .= 'filt_lib[\''.$soname.'\']=new Array(""';
		foreach my $row_n ( keys %{$filt_library->{$soname}} ) {
			$res .= ','.$row_n;
		}
		$res .= ');'.$nl;
	}
	
	# javascript: filt_com
	$res .= 'var filt_com = Array();'.$nl;
	foreach my $component ( keys %$filt_component ) {
		$res .= 'filt_com[\''.$component.'\']=new Array(""';
		foreach my $row_n ( keys %{$filt_component->{$component}} ) {
			$res .= ','.$row_n;
		}
		$res .= ');'.$nl;
	}
	
	$res .= <<HEREDOC;

// Convert filt_lib into a hash map
for ( var lib in filt_lib ) {
	var hash = new Array();
	for ( var row_n in filt_lib[lib] ) {
		hash[filt_lib[lib][row_n]] = 1;
	}
	filt_lib[lib] = hash;
}

// Convert filt_com into a hash map
for ( var com in filt_com ) {
	var hash = new Array();
	for ( var row_n in filt_com[com] ) {
		hash[filt_com[com][row_n]] = 1;
	}
	filt_com[com] = hash;
}

// Convert filt_show into a hash map
for ( var what in filt_show ) {
	var hash = new Array();
	for ( var row_n in filt_show[what] ) {
		hash[filt_show[what][row_n]] = 1;
	}
	filt_show[what] = hash;
}

function highlight_select( obj )
{
	if ( obj.options ) {
		for ( var i in obj.options ) {
			if ( obj.options[i] && obj.options[i].style ) {
				obj.options[i].style.backgroundColor = '#fff';
			}
		}
		if ( obj.selectedIndex ) {
			obj.options[obj.selectedIndex].style.backgroundColor = '#AAFFAA';
		}
	}
}

function apply_filter()
{
	var filt_obj = document.getElementById('filt');
	var filt = filt_obj.value;
	highlight_select(filt_obj);
	
	var com_select_obj = document.getElementById('com_select');
	var com_select = com_select_obj.value;
	highlight_select(com_select_obj);
	
	var lib_select_obj = document.getElementById('lib_select');
	var lib_select = lib_select_obj.value;
	highlight_select(lib_select_obj);

	var new_i_table = l[0]; // header line
	var shown = 0;
	
	for ( var i=1; i < l.length; i++ ) {
		if ( lib_select && (!filt_lib[lib_select] || !filt_lib[lib_select][i]) ) continue;
		if ( com_select && (!filt_com[com_select] || !filt_com[com_select][i]) ) continue;
		if ( filt && (!filt_show[filt] || !filt_show[filt][i]) ) continue;
		new_i_table += l[i];
		shown++;
	}
	
	var total = l.length - 1;
	lcp.innerHTML = shown + (shown != total ? ' of '+total : '') +' interface'+(total!=1 ? 's' : '')+' shown.';
	
	wrapper.innerHTML = '$table_open' + new_i_table + '</table>';
	return false;
}

// ]]>
</script>
HEREDOC

	if ( $_GET{'show'} || $_GET{'show_lib'} ) {
		my $filt_value = $_GET{'show'};
		my $filt_lib_value = $_GET{'show_lib'};
		
		$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
HEREDOC
		if ( $filt_value ) {
			$res .= <<HEREDOC;
	var filt = document.getElementById('filt');
	filt.value = '$filt_value';
HEREDOC
		}
		if ( $filt_lib_value ) {
			$res .= <<HEREDOC;
	var lib_select = document.getElementById('lib_select');
	lib_select.value = '$filt_lib_value';
HEREDOC
		}
		
		$res .= <<HEREDOC;
	apply_filter();
// ]]>
</script>
HEREDOC
	}

	return $res;
}

sub int_details {
	my ($Iname, $ver, $selected_lsb_ver) = @_;
	
	my $int_obj = $Analysis::ints->{$Iname}{$ver};
	
	if ( !defined $int_obj ) {
		return "Error: There is no such interface: '$Iname' '$ver'";
	}
	
	my $unmangled = ( $int_obj->{UNMANGLED} or $Iname );
	
	$res = "";
	my $nl = "";
	
	# Reference to LSB Navigator:
	my $navi_link = "http://www.linuxfoundation.org/navigator/browse/int_single.php"
			."?cmd=list-by-name&amp;Iname=".$Iname;
	
	# Title
	$res .= '<h1>Interface: <a href="'.$navi_link.'" target="_blank">'.$unmangled.'</a>'
			.($ver ? " ($ver)" : '').'</h1>';
	
	$res .= '<hr />'.$nl;
	
	# TODO: Int KB info
	
	# Check IntAttr
	if ( $int_obj->{ATTR} ) {
		foreach my $type ( sort keys %{$int_obj->{ATTR}} ) {
			my $attr = "";
			my $bgcolor = '#fec';
			if ( $type eq "Recommendation" ) {
				$attr .= '<h3>LSB Recommendation</h3>';
				$attr .= $int_obj->{ATTR}->{$type};
			}
			elsif ( $type eq "Rejection" ) {
				$attr .= '<h3>LSB Rejection</h3>';
				$attr .= $int_obj->{ATTR}->{$type};
			}
			else { # Custom
				$attr .= $int_obj->{ATTR}->{$type};
			}
			$attr =~ s/^\s+//; # remove unnecessary spaces
			$attr =~ s/\s+$//; #
			$attr .= "." if $attr =~ /[^\s\.]\s*$/;
			
			$res .= '<div class="box" style="background:'.$bgcolor.'">'.$attr.'</div>';
		}

		$res .= '<hr />'.$nl;
	}

	# LSB Status
	$res .= '<h3>LSB Status:</h3>'.$nl;

	my ($lsb_status, $lsb_status_ver) = int_lsb_status($int_obj, $selected_lsb_ver);
	
	if ( $lsb_status eq "NOT" ) {
		$res .= '<div style="color:red">Never been in LSB.</div>'.$nl;
	}
	elsif ( $lsb_status eq "FUTURE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. (Appeared in LSB '.$lsb_status_ver.')</div>'.$nl;
	}
	elsif ( $lsb_status eq "CANDIDATE" ) {
		$res .= '<div style="color:red">Not in LSB. Candidate for LSB '.$lsb_status_ver.'</div>'.$nl;
	}
	elsif ( $lsb_status eq "OPTIONAL" ) {
		$res .= '<div style="color:blue">Optional since LSB '.$lsb_status_ver.'</div>'.$nl;
	}
	elsif ( $lsb_status eq "IN" ) {
		$res .= '<div style="color:green">Mandatory since LSB '.$lsb_status_ver.'</div>'.$nl;
	}
	elsif ( $lsb_status eq "DEPRECATED" ) {
		$res .= '<div style="color:blue">Deprecated since LSB '.$lsb_status_ver.'</div>'.$nl;
	}
	elsif ( $lsb_status eq "WITHDRAWN" ) {
		$res .= '<div style="color:red">Withdrawn since LSB '.$lsb_status_ver.'</div>'.$nl;
	}

	$res .= '<hr />'.$nl;
	
	if ( $int_obj->{LIBRARIES} ) {
		# In Libraries
		$res .= '<h2>In Libraries</h2>';
		
		$res .= '<ul>';
		foreach my $soname ( @{$int_obj->{LIBRARIES}} ) {
			my $soname_link = href( tab_link('libs').'&rel=1&lib='.$soname );
			if (   $Analysis::libs->{$soname} && !$Analysis::libs->{$soname}{APPROVED} 
				or $Analysis::libs_impl->{$soname} && !$Analysis::libs_impl->{$soname}{APPROVED} )
			{
				$soname = '<a href="'.html($soname_link).'" style="color:#907000">'.html($soname).'</a>';
			}
			else
			{
				$soname = '<a href="'.html($soname_link).'">'.html($soname).'</a>';
			}
			$res .= '<li>'.$soname.'</li>';
		}
		$res .= '</ul>';
		
		$res .= '<hr />'.$nl;
	}
		
	# Presence in Distributions
	$res .= '<h2>Presence in Distributions</h2>';
	
	my %present = ();
	my %missed = ();
	foreach my $Did ( keys %$Analysis::all_distrs ) {
		my $distr_name = distr_name( $Analysis::all_distrs->{$Did} );
		if ( $int_obj->{DISTRS}{$Did} ) {
			$present{$distr_name} = $Did;
		} else {
			$missed{$distr_name} = $Did;
		}
	}
	
	if ( !%present && $int_obj->{UNAPPROVED} ) {
		%missed = ();
		$res .= '<p><i>No information.</i></p>';
	}
	
	if ( %present ) {
		$res .= '<h3>Present in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %present ) {
			my $Did = $present{$distr_name};
			
			if ( 0 ) { # Do not delete this block!
				my $soname_linked = "";
				foreach my $lib_with_path ( split /,\s*/, $int_obj->{DISTRS}{$Did} ) {
					if ( $soname_linked ) { $soname_linked .= ', '; }
					if ( $lib_with_path =~ /^(.*)\/([^\/]+)$/ ) { # Link to lib info
						my ($path, $soname) = ($1, $2);
						my $lib_link = href( tab_link('libs').'&lib='.$soname );
						$soname_linked .= $path.'/'.'<a style="color:blue" href="'.html($lib_link).'">'.html($soname).'</a>';
					} else {
						$soname_linked .= $lib_with_path;
					}
				}
			}

			my $distr_link = href( tab_link('distrs').'&id='.$Did );
						
			$res .= '<li><a style="color:green" href="'.html($distr_link).'">'
						.html($distr_name).'</a></li>';
		}
		$res .= '</ul>';
	}
	
	if ( %missed ) {
		$res .= '<h3>Missing in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %missed ) {
			my $Did = $missed{$distr_name};
			
			my $distr_link = href( tab_link('distrs').'&id='.$Did );
			
			$res .= '<li><a style="color:red" href="'.html($distr_link).'">'
						.html($distr_name).'</a></li>';
		}
		$res .= '</ul>';
	}
	
	$res .= '<hr />'.$nl;
	
	return $res;
}

sub tab_link {
	my ($tab) = @_;
	
	return '?details='.$_GET{'details'}.'&tab='.$tab;
}

sub href {
	my ($link) = @_;
	
	$link =~ s/\+/%2B/g;
	$link =~ s/ /%20/g;
	
	return $link;
}

sub html {
	my ($txt) = @_;
	
	$txt =~ s/&/&amp;/g;
	$txt =~ s/</&lt;/g;
	$txt =~ s/>/&gt;/g;
	
	return $txt;
}

sub html_arg {
	my ($txt) = @_;
	
	$txt =~ s/&/&amp;/g;
	$txt =~ s/"/&quot;/g;
	
	return $txt;
}
